from time import perf_counter

import numpy as np
import pytest
import scipy.linalg as sla
import scipy.stats as st
from scipy.special import digamma, gammaln

from skchange.change_detectors import PELT, MovingWindow
from skchange.costs import MultivariateTCost
from skchange.costs._multivariate_t_cost import (
    _isotropic_mv_t_dof_estimate,
    _iterative_mv_t_dof_estimate,
    _kurtosis_mv_t_dof_estimate,
    _loo_iterative_mv_t_dof_estimate,
    _multivariate_t_log_likelihood,
    _solve_for_mle_scale_matrix,
    maximum_likelihood_mv_t_scale_matrix,
)
from skchange.utils.numba import numba_available


def estimate_scale_matrix_trace_nojit(centered_samples: np.ndarray, dof: float):
    """Estimate the scale parameter of the MLE covariance matrix."""
    p = centered_samples.shape[1]
    squared_norms = np.sum(centered_samples * centered_samples, axis=1)
    z_bar = np.log(squared_norms[squared_norms > 1.0e-12]).mean()
    log_alpha = z_bar - np.log(dof) + digamma(0.5 * dof) - digamma(p / 2.0)
    return p * np.exp(log_alpha)


def initial_scale_matrix_estimate_nojit(
    centered_samples: np.ndarray,
    dof: float,
    num_zeroed_samples: int = 0,
    apply_trace_correction: bool = True,
):
    """Estimate the scale matrix given centered samples and degrees of freedom."""
    n, p = centered_samples.shape
    num_effective_samples = n - num_zeroed_samples

    sample_covariance_matrix = (
        centered_samples.T @ centered_samples
    ) / num_effective_samples

    if apply_trace_correction:
        scale_trace_estimate = estimate_scale_matrix_trace_nojit(centered_samples, dof)
        sample_covariance_matrix *= scale_trace_estimate / np.trace(
            sample_covariance_matrix
        )

    return sample_covariance_matrix


def scale_matrix_fixed_point_iteration(
    scale_matrix: np.ndarray,
    dof: float,
    centered_samples: np.ndarray,
    num_zeroed_samples: int = 0,
):
    """Compute the MLE covariance residual for a mv_t distribution."""
    n, p = centered_samples.shape

    # Subtract the number of 'zeroed' samples:
    effective_num_samples = n - num_zeroed_samples

    inv_cov_2d = sla.solve(
        scale_matrix, np.eye(p), assume_a="pos", overwrite_a=False, overwrite_b=True
    )
    z_scores = np.einsum("ij,jk,ik->i", centered_samples, inv_cov_2d, centered_samples)

    sample_weight = (p + dof) / (dof + z_scores)
    weighted_samples = centered_samples * sample_weight[:, np.newaxis]

    reconstructed_scale_matrix = (
        weighted_samples.T @ centered_samples
    ) / effective_num_samples

    return reconstructed_scale_matrix


def solve_mle_scale_matrix_nojit(
    initial_scale_matrix: np.ndarray,
    centered_samples: np.ndarray,
    dof: float,
    num_zeroed_samples: int = 0,
    max_iter: int = 50,
    abs_tol: float = 1.0e-3,
) -> np.ndarray:
    """Perform fixed point iterations for the MLE scale matrix."""
    scale_matrix = initial_scale_matrix.copy()
    temp_cov_matrix = initial_scale_matrix.copy()

    # Compute the MLE covariance matrix using fixed point iteration:
    for iteration in range(max_iter):
        temp_cov_matrix = scale_matrix_fixed_point_iteration(
            scale_matrix=scale_matrix,
            dof=dof,
            centered_samples=centered_samples,
            num_zeroed_samples=num_zeroed_samples,
        )

        # Note: 'ord = None' computes the Frobenius norm.
        residual = np.linalg.norm(temp_cov_matrix - scale_matrix, ord=None)

        scale_matrix = temp_cov_matrix.copy()
        if residual < abs_tol:
            break

    return scale_matrix, iteration


def maximum_likelihood_scale_matrix_nojit(
    centered_samples: np.ndarray,
    dof: float,
    abs_tol: float = 1.0e-3,
    max_iter: int = 50,
    num_zeroed_samples: int = 0,
    initial_trace_correction: bool = True,
) -> np.ndarray:
    """Compute the MLE scale matrix for a multivariate t-distribution.

    Parameters
    ----------
    centered_samples : np.ndarray
        The centered samples from the multivariate t-distribution.
    dof : float
        The degrees of freedom of the multivariate t-distribution.

    Returns
    -------
    np.ndarray
        The MLE covariance matrix of the multivariate t-distribution.
    """
    # Initialize the scale matrix:
    mle_scale_matrix = initial_scale_matrix_estimate_nojit(
        centered_samples,
        dof,
        num_zeroed_samples=num_zeroed_samples,
        apply_trace_correction=initial_trace_correction,
    )

    mle_scale_matrix, inner_iterations = solve_mle_scale_matrix_nojit(
        initial_scale_matrix=mle_scale_matrix,
        centered_samples=centered_samples,
        dof=dof,
        num_zeroed_samples=num_zeroed_samples,
        max_iter=max_iter,
        abs_tol=abs_tol,
    )

    return mle_scale_matrix


def approximate_mv_t_scale_matrix_gradient(
    scale_matrix: np.ndarray,
    centered_samples: np.ndarray,
    dof: float,
    epsilon: float = 1e-4,
):
    """
    Approximate the gradient of the scale matrix.

    Parameters
    ----------
    centered_samples : `np.ndarray`
        Centered samples.
    scale_matrix : `np.ndarray`
        Scale matrix.
    epsilon : `float`, optional (default=1e-4)
        Epsilon.

    Returns
    -------
    grad : `np.ndarray`
        Gradient.
    """
    p = centered_samples.shape[1]
    grad = np.zeros((p, p))

    for i in range(p):
        for j in range(p):
            scale_matrix_plus = scale_matrix.copy()
            scale_matrix_plus[i, j] += epsilon

            scale_matrix_minus = scale_matrix.copy()
            scale_matrix_minus[i, j] -= epsilon

            ll_plus = st.multivariate_t.logpdf(
                centered_samples, loc=np.zeros(p), shape=scale_matrix_plus, df=dof
            ).sum()
            ll_minus = st.multivariate_t.logpdf(
                centered_samples, loc=np.zeros(p), shape=scale_matrix_minus, df=dof
            ).sum()

            grad[i, j] = (ll_plus - ll_minus) / (2 * epsilon)

    return grad


def _multivariate_t_log_likelihood_like_scipy(
    scale_matrix, centered_samples: np.ndarray, dof: float
) -> float:
    """Calculate the log likelihood of a multivariate t-distribution.

    Directly from the definition of the multivariate t-distribution.
    Mirroring the scipy implementation of the multivariate t-distribution.
    Source: https://en.wikipedia.org/wiki/Multivariate_t-distribution

    Parameters
    ----------
    scale_matrix : np.ndarray
        The scale matrix of the multivariate t-distribution.
    centered_samples : np.ndarray
        The centered samples from the multivariate t-distribution.
    dof : float
        The degrees of freedom of the multivariate t-distribution.

    Returns
    -------
    float
        The log likelihood of the multivariate t-distribution.
    """
    num_samples, sample_dim = centered_samples.shape

    # Compute the log likelihood of the segment using numba, but similar to scipy:
    scale_spectrum, scale_eigvecs = np.linalg.eigh(scale_matrix)
    log_det_scale_matrix = np.sum(np.log(scale_spectrum))

    scale_sqrt_inv = np.multiply(scale_eigvecs, 1.0 / np.sqrt(scale_spectrum))
    mahalonobis_distances = np.square(np.dot(centered_samples, scale_sqrt_inv)).sum(
        axis=-1
    )

    # Normalization constants:
    exponent = 0.5 * (dof + sample_dim)
    A = gammaln(exponent)
    B = gammaln(0.5 * dof)
    C = 0.5 * sample_dim * np.log(dof * np.pi)
    D = 0.5 * log_det_scale_matrix
    normalization_contribution = num_samples * (A - B - C - D)

    sample_contributions = -exponent * np.log1p(mahalonobis_distances / dof)

    total_log_likelihood = normalization_contribution + sample_contributions.sum()

    return total_log_likelihood


def test_mv_t_log_likelihood(seed=4125, num_samples=100, p=8, t_dof=5.0):
    # TODO: Parametrize and test over wide range of input values.
    np.random.seed(seed)

    mv_t_samples = st.multivariate_t(loc=np.zeros(p), shape=np.eye(p), df=t_dof).rvs(
        num_samples
    )

    sample_medians = np.median(mv_t_samples, axis=0)
    X_centered = mv_t_samples - sample_medians
    mle_scale_matrix = maximum_likelihood_mv_t_scale_matrix(
        X_centered, t_dof, abs_tol=1.0e-6, rel_tol=0.0, max_iter=50
    )

    ll_scipy = (
        st.multivariate_t(loc=sample_medians, shape=mle_scale_matrix, df=t_dof)
        .logpdf(mv_t_samples)
        .sum()
    )

    ll_manual_scipy = _multivariate_t_log_likelihood_like_scipy(
        scale_matrix=mle_scale_matrix, centered_samples=X_centered, dof=t_dof
    )

    ll_numba = _multivariate_t_log_likelihood(
        scale_matrix=mle_scale_matrix, centered_samples=X_centered, dof=t_dof
    )
    ll_differences = np.diff(np.array([ll_scipy, ll_manual_scipy, ll_numba, ll_scipy]))

    np.testing.assert_allclose(ll_differences, 0, atol=1e-4)


def test_scale_matrix_mle(seed=4125):
    """Test scale matrix MLE."""
    # TODO: Parametrize and test over wide range of input values.
    np.random.seed(seed)
    n_samples = 50
    p = 3
    t_dof = 5.0

    random_nudge = np.random.randn(p).reshape(-1, 1)
    true_scale_matrix = np.eye(p) + 0.5 * random_nudge @ random_nudge.T

    true_mean = np.arange(p) * (-1 * np.ones(p)).cumprod()

    mv_t_samples = st.multivariate_t(
        loc=true_mean, shape=true_scale_matrix, df=t_dof
    ).rvs(n_samples)

    sample_medians = np.median(mv_t_samples, axis=0)
    centered_samples = mv_t_samples - sample_medians

    # Estimate the scale matrix:
    mle_scale_matrix = maximum_likelihood_scale_matrix_nojit(
        centered_samples, t_dof, abs_tol=1.0e-6
    )

    # Compute approximate gradients:
    mle_scale_matrix_grad = approximate_mv_t_scale_matrix_gradient(
        mle_scale_matrix, centered_samples, t_dof
    )
    true_scale_matrix_grad = approximate_mv_t_scale_matrix_gradient(
        true_scale_matrix, centered_samples, t_dof
    )

    # Assure that the MLE scale matrix gradient is close to zero:
    np.testing.assert_allclose(mle_scale_matrix_grad, 0.0, atol=1e-5)

    # Assure that the norm of the gradient for the true scale matrix is larger:
    assert np.linalg.norm(true_scale_matrix_grad) > np.linalg.norm(
        mle_scale_matrix_grad
    ), "True scale matrix gradient is not larger than the MLE gradient."

    # Assure that we've increased the log-likelihood with the MLE scale matrix:
    true_scale_matrix_ll = st.multivariate_t.logpdf(
        centered_samples, loc=np.zeros(p), shape=true_scale_matrix, df=t_dof
    ).sum()

    mle_scale_matrix_ll = st.multivariate_t.logpdf(
        centered_samples, loc=np.zeros(p), shape=mle_scale_matrix, df=t_dof
    ).sum()

    assert (
        mle_scale_matrix_ll > true_scale_matrix_ll
    ), "MLE log-likelihood is not maximal."


def test_loo_scale_matrix_mle(seed=4125):
    """Test leave-one-out scale matrix MLE."""
    np.random.seed(seed)
    n_samples = 50
    p = 3
    t_dof = 5.0
    mle_scale_abs_tol = 1.0e-6

    # Test the leave-one-out MLE scale matrix on a subset of the samples:
    num_test_indices = 10

    random_nudge = np.random.randn(p).reshape(-1, 1)
    true_scale_matrix = np.eye(p) + 0.5 * random_nudge @ random_nudge.T

    true_mean = np.arange(p) * (-1 * np.ones(p)).cumprod()

    mv_t_samples = st.multivariate_t(
        loc=true_mean, shape=true_scale_matrix, df=t_dof
    ).rvs(n_samples)

    sample_medians = np.median(mv_t_samples, axis=0)
    centered_samples = mv_t_samples - sample_medians

    loo_indices = np.random.choice(n_samples, num_test_indices, replace=False)
    for loo_index in loo_indices:
        loo_centered_samples = np.delete(centered_samples, loo_index, axis=0)
        direct_loo_mle_scale = maximum_likelihood_mv_t_scale_matrix(
            loo_centered_samples,
            t_dof,
            abs_tol=mle_scale_abs_tol,
            rel_tol=0.0,
            max_iter=100,
        )

        index_loo_mle_scale = maximum_likelihood_mv_t_scale_matrix(
            centered_samples,
            t_dof,
            loo_index=loo_index,
            abs_tol=mle_scale_abs_tol,
            rel_tol=0.0,
            max_iter=100,
        )

        np.testing.assert_allclose(
            direct_loo_mle_scale, index_loo_mle_scale, atol=1e-16
        )


def test_scale_matrix_numba_benchmark(
    n_trials=10,
    n_samples=1_000,
    p=3,
    t_dof=5.0,
    initial_trace_correction=True,
    verbose=False,
):
    """Benchmark numba vs non-numba scale matrix computation."""
    if not numba_available:
        pytest.skip("Numba not available, cannot test benchmark.")

    times_njit = []
    times_normal = []

    for seed in range(n_trials):
        np.random.seed(seed)

        # Generate test data
        random_nudge = np.random.randn(p).reshape(-1, 1)
        true_scale_matrix = np.eye(p) + 0.5 * random_nudge @ random_nudge.T
        true_mean = np.arange(p) * (-1 * np.ones(p)).cumprod()
        mv_t_samples = st.multivariate_t(
            loc=true_mean, shape=true_scale_matrix, df=t_dof
        ).rvs(n_samples)
        centered_samples = mv_t_samples - np.median(mv_t_samples, axis=0)

        if seed == 0:
            # Ensure compilation time is not measured:
            maximum_likelihood_mv_t_scale_matrix(
                centered_samples,
                t_dof,
                abs_tol=1.0e-9,
                rel_tol=0.0,
                max_iter=100,
            )

        # Time numba version
        start = perf_counter()
        numba_mle_scale_matrix = maximum_likelihood_mv_t_scale_matrix(
            centered_samples,
            t_dof,
            abs_tol=1.0e-9,
            rel_tol=0.0,
            max_iter=100,
        )
        end = perf_counter()
        times_njit.append(end - start)

        # Time nojit version
        start = perf_counter()
        nojit_mle_scale_matrix = maximum_likelihood_scale_matrix_nojit(
            centered_samples,
            t_dof,
            initial_trace_correction=initial_trace_correction,
            abs_tol=1.0e-9,
        )
        end = perf_counter()
        times_normal.append(end - start)

        # Assert numba version is close to normal version:
        np.testing.assert_allclose(
            numba_mle_scale_matrix, nojit_mle_scale_matrix, atol=1e-9
        )

    # Assert numba version is faster on average:
    mean_numba_time = np.mean(times_njit)
    mean_normal_time = np.mean(times_normal)
    numba_speedup = mean_normal_time / mean_numba_time

    if verbose:
        print(f"Mean time normal: {mean_normal_time:.3e}")
        print(f"Mean time numba: {mean_numba_time:.3e}")
        print(f"Numba speedup: {numba_speedup:.3f}")

    assert numba_speedup > 1, "Numba version should be faster"


def test_isotropic_and_kurtosis_t_dof_estimates():
    seed = 4125
    n_samples = 1000
    p = 5
    t_dof = 5.0

    np.random.seed(seed)

    # Spd covariance matrix:
    random_nudge = np.random.randn(p).reshape(-1, 1)
    cov = np.eye(p) + 0.5 * random_nudge @ random_nudge.T

    mean = np.arange(p) * (-1 * np.ones(p)).cumprod()

    mv_t_dist = st.multivariate_t(loc=mean, shape=cov, df=t_dof)
    mv_t_samples = mv_t_dist.rvs(n_samples)

    sample_medians = np.median(mv_t_samples, axis=0)
    centered_samples = mv_t_samples - sample_medians

    # Test isotropic estimate:
    isotropic_dof = _isotropic_mv_t_dof_estimate(
        centered_samples, infinite_dof_threshold=50.0
    )
    assert isotropic_dof > 0, "Isotropic dof estimate should be positive."
    assert np.abs(isotropic_dof - t_dof) < 1.0, "Isotropic dof estimate is off."

    # Test kurtosis estimate:
    kurtosis_dof = _kurtosis_mv_t_dof_estimate(
        centered_samples, infinite_dof_threshold=50.0
    )
    assert kurtosis_dof > 0, "Kurtosis dof estimate should be positive."
    assert np.abs(kurtosis_dof - t_dof) < 1.0, "Kurtosis dof estimate is off."


def test_iso_and_kurt_dof_estimates_on_gaussian_data():
    seed = 0
    np.random.seed(seed)
    n_samples = 500
    p = 2
    t_dof = 5.0

    mv_t_samples = st.multivariate_t.rvs(
        df=t_dof, loc=np.zeros(p), shape=np.eye(p), size=n_samples
    )

    mv_normal_samples = st.multivariate_normal.rvs(
        mean=np.zeros(p), cov=np.eye(p), size=n_samples
    )

    mv_t_kurt_dof_est = _kurtosis_mv_t_dof_estimate(
        mv_t_samples, infinite_dof_threshold=50.0
    )
    assert np.isfinite(mv_t_kurt_dof_est) and (
        mv_t_kurt_dof_est > 0.0
    ), "Kurtosis dof estimate should be finite on multivariate t samples."

    mv_t_isotropic_dof_est = _isotropic_mv_t_dof_estimate(
        mv_t_samples, infinite_dof_threshold=50.0
    )
    assert np.isfinite(mv_t_isotropic_dof_est) and (
        mv_t_isotropic_dof_est > 0.0
    ), "Isotropic dof estimate should be finite on multivariate t samples."

    normal_kurt_dof_est = _kurtosis_mv_t_dof_estimate(
        mv_normal_samples, infinite_dof_threshold=50.0
    )
    assert np.isposinf(
        normal_kurt_dof_est
    ), "Kurtosis dof estimate should be infinite on Gaussian data."

    normal_isotropic_dof_est = _isotropic_mv_t_dof_estimate(
        mv_normal_samples, infinite_dof_threshold=50.0
    )
    assert np.isposinf(
        normal_isotropic_dof_est
    ), "Isotropic dof estimate should be infinite on Gaussian data."


def test_iterative_t_dof_estimate():
    seed = 4125
    n_samples = 150
    p = 5
    t_dof = 5.0

    np.random.seed(seed)

    # Spd covariance matrix:
    random_nudge = np.random.randn(p).reshape(-1, 1)
    cov = np.eye(p) + 0.5 * random_nudge @ random_nudge.T

    mean = np.arange(p) * (-1 * np.ones(p)).cumprod()

    mv_t_dist = st.multivariate_t(loc=mean, shape=cov, df=t_dof)
    mv_t_samples = mv_t_dist.rvs(n_samples)

    sample_medians = np.median(mv_t_samples, axis=0)
    centered_samples = mv_t_samples - sample_medians

    # Test data-driven estimate:
    initial_t_dof_estimate = 10.0
    iterative_dof_estimate = _iterative_mv_t_dof_estimate(
        centered_samples,
        initial_dof=initial_t_dof_estimate,
        infinite_dof_threshold=50.0,
        dof_max_iter=100,
        mle_scale_abs_tol=1.0e-6,
        mle_scale_rel_tol=0.0,
        mle_scale_max_iter=100,
    )
    assert iterative_dof_estimate > 0, "Data-driven dof estimate should be positive."
    assert (
        np.abs(iterative_dof_estimate - t_dof) < 1.0
    ), "Data-driven dof estimate is off."


def test_loo_iterative_t_dof_estimate():
    seed = 4125
    n_samples = 200
    p = 5
    t_dof = 5.0

    np.random.seed(seed)

    # Spd covariance matrix:
    random_nudge = np.random.randn(p).reshape(-1, 1)
    cov = np.eye(p) + 0.5 * random_nudge @ random_nudge.T

    mean = np.arange(p) * (-1 * np.ones(p)).cumprod()

    mv_t_dist = st.multivariate_t(loc=mean, shape=cov, df=t_dof)
    mv_t_samples = mv_t_dist.rvs(n_samples)

    sample_medians = np.median(mv_t_samples, axis=0)
    centered_samples = mv_t_samples - sample_medians

    # Test data-driven estimate:
    initial_t_dof_estimate = 10.0
    loo_iterative_dof = _loo_iterative_mv_t_dof_estimate(
        centered_samples,
        initial_dof=initial_t_dof_estimate,
        infinite_dof_threshold=50.0,
        mle_scale_abs_tol=1.0e-6,
        mle_scale_rel_tol=0.0,
        mle_scale_max_iter=100,
    )
    assert loo_iterative_dof > 0, "LOO data-driven dof estimate should be positive."
    assert (
        np.abs(loo_iterative_dof - t_dof) < 0.15
    ), "LOO data-driven dof estimate is off."


def test_iterative_dof_estimate_returns_inf_on_gaussian_data():
    seed = 4125
    n_samples = 150
    p = 5

    np.random.seed(seed)

    mean = np.arange(p) * (-1 * np.ones(p)).cumprod()
    cov = np.eye(p)

    mv_normal_samples = st.multivariate_normal(mean=mean, cov=cov).rvs(n_samples)

    sample_medians = np.median(mv_normal_samples, axis=0)
    centered_samples = mv_normal_samples - sample_medians

    # Test data-driven estimate:
    initial_t_dof_estimate = 4.0
    iterative_dof_estimate = _iterative_mv_t_dof_estimate(
        centered_samples,
        initial_dof=initial_t_dof_estimate,
        infinite_dof_threshold=50.0,
        dof_max_iter=100,
        mle_scale_abs_tol=1.0e-6,
        mle_scale_rel_tol=0.0,
        mle_scale_max_iter=100,
    )

    assert np.isposinf(
        iterative_dof_estimate
    ), "Dof estimate should be infinite on Gaussian data."


def test_loo_iterative_dof_estimate_returns_inf_on_gaussian_data():
    seed = 4125
    n_samples = 500
    p = 5

    np.random.seed(seed)

    mean = np.arange(p) * (-1 * np.ones(p)).cumprod()
    cov = np.eye(p)

    mv_normal_samples = st.multivariate_normal(mean=mean, cov=cov).rvs(n_samples)

    sample_medians = np.median(mv_normal_samples, axis=0)
    centered_samples = mv_normal_samples - sample_medians

    initial_t_dof_estimate = 4.0
    loo_iterative_dof = _loo_iterative_mv_t_dof_estimate(
        centered_samples,
        initial_dof=initial_t_dof_estimate,
        infinite_dof_threshold=50.0,
        dof_max_iter=100,
        mle_scale_abs_tol=1.0e-6,
        mle_scale_rel_tol=0.0,
        mle_scale_max_iter=100,
    )

    assert np.isposinf(
        loo_iterative_dof
    ), "Dof estimate should be infinite on Gaussian data."


def test_MultiVariateTCost_with_PELT(
    seed=5212,
    n_samples=20,
    p=2,
    t_dof=5.0,
    cost_dof=None,
    mle_scale_abs_tol=1.0e-3,
    mle_scale_rel_tol=1.0e-2,
):
    np.random.seed(seed)

    mean_1 = np.arange(p) * (-1 * np.ones(p)).cumprod()
    scale_1 = np.eye(p)

    mv_t_1_samples = st.multivariate_t(loc=mean_1, shape=scale_1, df=t_dof).rvs(
        n_samples
    )

    mean_2 = mean_1 + 1.0
    scale_2 = np.eye(p)
    scale_2[0, 0] = 10.0
    mv_t_2_samples = st.multivariate_t(loc=mean_2, shape=scale_2, df=t_dof).rvs(
        n_samples
    )

    X = np.vstack([mv_t_1_samples, mv_t_2_samples])

    mv_t_cost = MultivariateTCost(
        fixed_dof=cost_dof,
        mle_scale_abs_tol=mle_scale_abs_tol,
        mle_scale_rel_tol=mle_scale_rel_tol,
    )
    change_detector = PELT(cost=mv_t_cost, min_segment_length=2 * p + 1)

    segmentation = change_detector.fit_transform(X)
    change_points = change_detector.dense_to_sparse(segmentation)

    fitted_dof = change_detector.fitted_cost.dof_

    print(f"Change points: {change_points}")
    print(f"Estimated dof: {fitted_dof}")

    assert len(change_points) == 1, "Only one change point should be detected."
    assert (
        change_points.loc[0, "ilocs"] == n_samples
    ), "Change point should be at the end of the first segment."
    assert np.isfinite(fitted_dof), "Fitted dof should be finite."


def test_MultiVariateTCost_with_moving_window(
    seed=5212, n_samples=200, p=5, t_dof=5.0, cost_dof=None
):
    np.random.seed(seed)

    mean_1 = np.arange(p) * (-1 * np.ones(p)).cumprod()
    scale_1 = np.eye(p)

    mv_t_1_samples = st.multivariate_t(loc=mean_1, shape=scale_1, df=t_dof).rvs(
        n_samples
    )

    mean_2 = mean_1 + 1.0
    scale_2 = np.eye(p)
    scale_2[0, 0] = 10.0
    mv_t_2_samples = st.multivariate_t(loc=mean_2, shape=scale_2, df=t_dof).rvs(
        n_samples
    )

    X = np.vstack([mv_t_1_samples, mv_t_2_samples])

    t_cost = MultivariateTCost(fixed_dof=cost_dof)
    change_detector = MovingWindow(change_score=t_cost, bandwidth=int(0.8 * n_samples))

    segmentation = change_detector.fit_transform(X)
    change_points = change_detector.dense_to_sparse(segmentation)

    fitted_dof = change_detector.fitted_score.score_.cost_.dof_

    print(f"Change points: {change_points}")
    print(f"Estimated dof: {fitted_dof}")

    assert len(change_points) == 1, "Only one change point should be detected."
    assert (
        change_points.loc[0, "ilocs"] == n_samples
    ), "Change point should be at the end of the first segment."
    assert np.isfinite(fitted_dof), "Fitted dof should be finite."


def test_min_size_not_fitted():
    """Test that min_size returns None when MultivariateTCost is not fitted."""
    cost = MultivariateTCost()
    assert cost.min_size is None, "min_size should be None when the cost is not fitted."


def test_setting_fixed_dof():
    """Test that min_size returns None when MultivariateTCost is not fitted."""
    cost = MultivariateTCost(fixed_dof=5.0)
    assert cost.fixed_dof == 5.0, "Fixed dof should be set to 5.0."

    cost.fit(np.random.randn(100, 5))
    assert cost.dof_ == 5.0, "Fitted dof should be finite."


def test_nan_dof_raises_value_error():
    """Test that constructing MvTCost with nan dof raises ValueError."""
    with pytest.raises(
        ValueError,
        match="fixed_dof must be in",
    ):
        MultivariateTCost(fixed_dof=np.nan)


def test_fit_raises_value_error_for_non_positive_definite_scale_matrix():
    """Test that fit raises ValueError for non-positive definite scale matrix."""
    non_positive_definite_matrix = np.array([[1, 2], [2, 1]])
    cost = MultivariateTCost(param=(np.zeros(2), non_positive_definite_matrix))
    X = np.random.randn(100, 2)

    with pytest.raises(
        ValueError, match="covariance matrix must be positive definite."
    ):
        cost.fit(X)


def test_iterative_mv_t_dof_estimate_returns_inf_for_high_initial_dof():
    """Test that _iterative_mv_t_dof_estimate returns np.inf for high initial dof."""
    centered_samples = np.random.randn(100, 5)
    initial_dof = 100.0
    infinite_dof_threshold = 50.0

    dof_estimate = _iterative_mv_t_dof_estimate(
        centered_samples=centered_samples,
        initial_dof=initial_dof,
        infinite_dof_threshold=infinite_dof_threshold,
        mle_scale_abs_tol=1.0e-6,
        mle_scale_rel_tol=0.0,
        mle_scale_max_iter=100,
    )

    assert np.isposinf(
        dof_estimate
    ), "Dof estimate should be infinite for high initial dof."


def test_multivariate_t_log_likelihood_returns_nan_for_non_pos_def_scale_matrix():
    """Test that log likelihood returns np.nan for non-pos. def. scale matrix."""
    non_positive_definite_matrix = np.array([[1, 2], [2, 1]], dtype=np.float64)
    centered_samples = np.random.randn(100, 2)
    dof = 5.0

    log_likelihood = _multivariate_t_log_likelihood(
        scale_matrix=non_positive_definite_matrix,
        centered_samples=centered_samples,
        dof=dof,
    )

    assert np.isnan(
        log_likelihood
    ), "Log likelihood should be np.nan for non-positive definite scale matrix."


def test_solve_for_mle_scale_matrix_throws_value_error_if_max_iter_reached():
    """Test that _solve_for_mle_scale_matrix throws if max_iter is reached."""
    centered_samples = np.random.randn(100, 5)
    initial_scale_matrix = np.eye(5)
    dof = 5.0
    max_iter = 1  # Set max_iter to a low value to force the error

    with pytest.raises(RuntimeError, match="Maximum number of iterations reached"):
        _solve_for_mle_scale_matrix(
            initial_scale_matrix=initial_scale_matrix,
            centered_samples=centered_samples,
            dof=dof,
            max_iter=max_iter,
            abs_tol=1.0e-3,
            rel_tol=1.0e-3,
        )


def test_fixed_params_cost_higher_than_optim_param():
    """Test that fixed params cost is higher than optim param."""
    np.random.seed(4125)
    n_samples = 1000
    p = 5
    t_dof = 5.0

    fixed_loc = np.zeros(p)
    fixed_shape = np.eye(p) + 0.25 * np.random.randn(p, p)
    fixed_shape = 0.5 * (fixed_shape + fixed_shape.T)
    mv_t_samples = st.multivariate_t(loc=fixed_loc, shape=fixed_shape, df=t_dof).rvs(
        n_samples
    )

    cost_eval_intervals = np.array(
        [[0, n_samples], [0, n_samples // 2], [n_samples // 2, n_samples]]
    )

    # Compute the cost with fixed params and estimated dof
    fixed_params = (fixed_loc, np.eye(p))
    fixed_cost = MultivariateTCost(param=fixed_params)
    fixed_cost.fit(mv_t_samples)
    fixed_cost_evals = fixed_cost.evaluate(cost_eval_intervals)

    # Compute the cost with estimated params and dof:
    optim_cost = MultivariateTCost()
    optim_cost.fit(mv_t_samples)
    optim_cost_evals = optim_cost.evaluate(cost_eval_intervals)

    assert np.all(optim_cost_evals < fixed_cost_evals), "Fixed cost should be higher."
    assert fixed_cost.dof_ == optim_cost.dof_, "Estimated dof should be identical."



"""The ESAC score for detecting changes in the mean of high-dimensional data."""

import numpy as np
from scipy.stats import norm

from ..base import BaseIntervalScorer
from ..utils.numba import njit
from ..utils.numba.stats import col_cumsum
from ..utils.validation.parameters import check_larger_than
from ._cusum import cusum_score


@njit
def transform_esac(
    cusum_scores: np.ndarray,
    a_s: np.ndarray,
    nu_s: np.ndarray,
    t_s: np.ndarray,
    threshold: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    r"""
        Compute ESAC scores from CUSUM scores.

        This function calculates the penalised score for the ESAC algorithm,
        as defined in Equation (6) in [1]_. The outputs are penalised CUSUM
        scores computed from the input CUSUM scores.

    Parameters
    ----------
            cusum_scores (np.ndarray): A 2D array where each row represents the
                CUSUM scores for a specific candidate change location.
            a_s (np.ndarray): A 1D array of hard threshold values. Correspond to
                a(t) as defined in Equation (4) in [1]_ for each t specified in t_s.
            nu_s (np.ndarray): A 1D array of mean-centering terms. Correspond to
                nu(t) as defined after Equation (4) in [1]_ for each t specified in t_s.
            t_s (np.ndarray): A 1D array of candidate sparsity values corresponding
                to the element in a_s and nu_s.
            threshold (np.ndarray): A 1D array of penalty values, corresponding to
                \gamma(t) in Equation (4) in [1]_, where t is as defined in t_s.

    Returns
    -------
            tuple[np.ndarray, np.ndarray]: A tuple containing:
                - output_scores (np.ndarray): A 2D array of computed ESAC scores. The
                    array has one column, where each element represents the maximum
                    score for each candidate change location.
                - sargmax (np.ndarray): A 1D array giving sparisty level at which the
                    maximum score was achieved for each candidate change location.

    References
    ----------
    .. [1] Per August Jarval Moen, Ingrid Kristine Glad, Martin Tveten. Efficient
       sparsity adaptive changepoint estimation. Electron. J. Statist. 18 (2)
       3975 - 4038, 2024. https://doi.org/10.1214/24-EJS2294.
    """
    num_levels = len(threshold)
    num_cusum_scores = len(cusum_scores)
    output_scores = np.zeros(num_cusum_scores, dtype=np.float64)
    sargmax = np.zeros(num_cusum_scores, dtype=np.int64)

    for i in range(num_cusum_scores):
        temp_max = -np.inf
        for j in range(num_levels):
            temp_vec = (cusum_scores[i])[np.abs(cusum_scores[i]) > a_s[j]]
            if len(temp_vec) > 0:
                temp = np.sum(temp_vec**2 - nu_s[j]) - threshold[j]
                if temp > temp_max:
                    temp_max = temp
                    sargmax[i] = t_s[j]

        output_scores[i] = temp_max

    return output_scores.reshape(-1, 1), sargmax


class ESACScore(BaseIntervalScorer):
    """The ESAC score for detecting changes in the mean of high-dimensional data.

    This is the sparsity adaptive penalised CUSUM score for a change in the mean.
    The ESAC score is a penalised version of the CUSUM score, where the CUSUM of
    each time series is thresholded, mean-centered and penalised by a sparsity
    dependant penalty. The score is defined in Equation (6) in [1]_.

    Parameters
    ----------
    threshold_dense : float, optional, default=1.5
        The leading constant in the penalty function taken as in (8) in [1]_ in the
        dense case where the candidate sparisity level t is greater than or equal to
        sqrt(p * log(n)).
    threshold_sparse : float, optional, default=1.0
        The leading constant in the penalty function taken as in (8) in [1]_ in the
        sparse case where the candidate sparisity level t is less than sqrt(p * log(n)).

    References
    ----------
    .. [1] Per August Jarval Moen, Ingrid Kristine Glad, Martin Tveten. Efficient
       sparsity adaptive changepoint estimation. Electron. J. Statist. 18 (2)
       3975 - 4038, 2024. https://doi.org/10.1214/24-EJS2294.
    """

    _tags = {
        "authors": ["peraugustmoen", "Tveten"],
        "maintainers": ["peraugustmoen", "Tveten"],
        "task": "change_score",
        "distribution_type": "Gaussian",
        "is_conditional": False,
        "is_aggregated": True,
        "is_penalised": True,
    }

    def __init__(
        self,
        threshold_dense=1.5,
        threshold_sparse=1.0,
    ):
        self.threshold_dense = threshold_dense
        self.threshold_sparse = threshold_sparse
        super().__init__()

        check_larger_than(0.0, self.threshold_dense, "threshold_dense")
        check_larger_than(0.0, self.threshold_sparse, "threshold_sparse")

    def _fit(self, X: np.ndarray, y=None):
        """Fit the change score evaluator.

        Parameters
        ----------
        X : np.ndarray
            Data to evaluate. Must be a 2D array.
        y : None
            Ignored. Included for API consistency by convention.

        Returns
        -------
        self :
            Reference to self.
        """
        self._sums = col_cumsum(X, init_zero=True)
        self.n = X.shape[0]
        self.p = X.shape[1]
        n = self.n
        p = self.p

        if p == 1:
            self.a_s = np.array([0.0])
            self.nu_s = np.array([1.0])
            self.t_s = np.array([1])
            self.threshold = np.array(
                [self.threshold_dense * (np.sqrt(p * np.log(n)) + np.log(n))]
            )
        else:
            max_s = min(np.sqrt(p * np.log(n)), p)

            # Generate log2ss from 0 to floor(log2(max_s))
            log2ss = np.arange(0, np.floor(np.log2(max_s)) + 1)

            # calculate candidate sparsity values
            ss = 2**log2ss

            # add p as candidate sparsity and reverse order
            ss = np.concatenate(([p], ss[::-1]))
            self.t_s = np.array(ss, dtype=float)
            ss = np.array(ss, dtype=float)

            # Initialize as array
            self.a_s = np.zeros_like(ss, dtype=float)
            self.a_s[1:] = np.sqrt(
                2 * np.log(np.exp(1) * p * 4 * np.log(n) / ss[1:] ** 2)
            )

            # Calculate nu_as using the following formula:
            # nu(a) = E(Z^2|Z^2 > a^2) = 1 + a * exp(log(phi(a)) - log(1 - Phi(a))),
            # where Z ~ N(0, 1) and phi is the standard normal density function,
            # and Phi is the standard normal cumulative distribution function.
            log_dnorm = norm.logpdf(self.a_s)
            log_pnorm_upper = norm.logsf(self.a_s)

            self.nu_s = 1 + self.a_s * np.exp(log_dnorm - log_pnorm_upper)

            self.threshold = np.zeros_like(ss, dtype=float)
            self.threshold[0] = self.threshold_dense * (
                np.sqrt(4 * p * np.log(n)) + 4 * np.log(n)
            )
            self.threshold[1:] = self.threshold_sparse * (
                ss[1:] * np.log(np.exp(1) * p * 4 * np.log(n) / ss[1:] ** 2)
                + 4 * np.log(n)
            )

        return self

    def _evaluate(self, cuts: np.ndarray):
        """Evaluate the change score for a split within an interval.

        Evaluates the score for `X[start:split]` vs. `X[split:end]` for each
        start, split, end in cuts.

        Parameters
        ----------
        cuts : np.ndarray
            A 2D array with three columns of integer locations.
            The first column is the ``start``, the second is the ``split``, and the
            third is the ``end`` of the interval to evaluate.
            The difference between subsets ``X[start:split]`` and ``X[split:end]`` is
            evaluated for each row in `cuts`.

        Returns
        -------
        scores : np.ndarray
            A 2D array of change scores. One row for each cut. The number of
            columns is 1 if the change score is inherently multivariate. The number of
            columns is equal to the number of columns in the input data if the score is
            univariate. In this case, each column represents the univariate score for
            the corresponding input data column.
        """
        starts = cuts[:, 0]
        splits = cuts[:, 1]
        ends = cuts[:, 2]
        cusum_scores = cusum_score(starts, ends, splits, self._sums)
        thresholded_cusum_scores, sargmaxes = transform_esac(
            cusum_scores, self.a_s, self.nu_s, self.t_s, self.threshold
        )
        self.sargmaxes = sargmaxes
        return thresholded_cusum_scores

    @property
    def min_size(self) -> int | None:
        """Minimum size of the interval to evaluate.

        The size of each interval is defined as ``cuts[i, 1] - cuts[i, 0]``.

        Returns
        -------
        int or None
            The minimum valid size of an interval to evaluate. If ``None``, it is
            unknown what the minimum size is. E.g., the scorer may need to be fitted
            first to determine the minimum size.
        """
        return 1

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for change scores.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {"threshold_dense": 1.5, "threshold_sparse": 1.0},
            {"threshold_dense": 2.0, "threshold_sparse": 2.0},
        ]
        return params



"""Test cases for the ContinuousLinearTrendScore class and related functions."""

import datetime
import re

import numpy as np
import pandas as pd
import pytest

from skchange.change_detectors import MovingWindow, SeededBinarySegmentation
from skchange.change_scores import ContinuousLinearTrendScore
from skchange.change_scores._continuous_linear_trend_score import (
    analytical_cont_piecewise_linear_trend_score,
    lin_reg_cont_piecewise_linear_trend_score,
)
from skchange.datasets import generate_continuous_piecewise_linear_signal


def test_moving_window_single_changepoint():
    """Test MovingWindow with ContinuousLinearTrendScore on a single changepoint."""
    # Generate data with a single changepoint at position 100
    true_change_points = [100]
    slopes = [0.5, -0.5]  # Positive slope followed by negative slope
    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=200,
        noise_std=1.5,
        random_seed=42,
    )

    # Create detector with ContinuousLinearTrendScore
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(),
        bandwidth=30,
        penalty=25,  # Tuned for this specific test case
    )

    # Fit and predict
    detected_cps = detector.fit_predict(df)

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Assert the changepoint is close to the true changepoint
    detected_cp = detected_cps.iloc[0, 0]
    cp_detection_margin = 3
    assert (
        abs(detected_cp - true_change_points[0]) <= cp_detection_margin
    ), f"Detected {detected_cp}, expected close to {true_change_points[0]}"


def test_moving_window_multiple_changepoints():
    """Test MovingWindow with ContinuousLinearTrendScore on multiple changepoints."""
    # Generate data with multiple changepoints
    true_change_points = [100, 200, 300]
    slopes = [0.5, -0.5, 0.4, -0.4]
    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=400,
        noise_std=1.5,
        random_seed=42,
    )

    # Create detector with ContinuousLinearTrendScore
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(), bandwidth=50, penalty=20
    )

    # Fit and predict
    detected_cps = detector.fit_predict(df)

    # Assert we found the correct number of changepoints
    assert len(detected_cps) == len(
        true_change_points
    ), f"Expected {len(true_change_points)} changepoints, found {len(detected_cps)}"

    # Assert the changepoints are close to the true changepoints
    cp_detection_margin = 3
    for i, cp in enumerate(detected_cps["ilocs"]):
        assert (
            abs(cp - true_change_points[i]) <= cp_detection_margin
        ), f"Detected {cp}, expected close to {true_change_points[i]}"


def test_seeded_binseg_single_changepoint():
    """Test SeededBinarySegmentation with ContinuousLinearTrendScore

    On a single changepoint.
    """
    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.1, -0.2]
    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=200,
        noise_std=1.5,
        random_seed=42,
    )

    # Create detector with ContinuousLinearTrendScore:
    detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(),
        penalty=25,
        selection_method="narrowest",
    )

    # Fit and predict
    detected_cps = detector.fit_predict(df)

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Assert the changepoint is close to the true changepoint
    detected_cp = detected_cps.iloc[0, 0]
    cp_detection_margin = 2
    assert (
        abs(detected_cp - true_change_points[0]) <= cp_detection_margin
    ), f"Detected {detected_cp}, expected close to {true_change_points[0]}"


def test_seeded_binseg_multiple_changepoints():
    """Test SeededBinarySegmentation with ContinuousLinearTrendScore

    On multiple changepoints.
    """
    # Generate data with multiple changepoints
    true_change_points = [100, 200, 300]
    slopes = [0.1, -0.2, 0.15, -0.1]
    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=400,
        noise_std=2.0,
        random_seed=42,
    )

    # Create detector with ContinuousLinearTrendScore:
    detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(),
        selection_method="narrowest",
        penalty=45.0,  # ilocs: 101, 195, 302
        # selection_method="greedy",
        ## Low penalty that finds 3 changepoints when using greedy:
        # penalty=45,
        ## High penalty that still finds 3 changepoints when using greedy:
        # penalty=500,
    )

    # Fit and predict:
    detected_cps = detector.fit_predict(df)

    # Assert we found the correct number of changepoints
    assert len(detected_cps) == len(
        true_change_points
    ), f"Expected {len(true_change_points)} changepoints, found {len(detected_cps)}"

    # Assert the changepoints are close to the true changepoints
    cp_detection_margin = 5
    for i, cp in enumerate(detected_cps["ilocs"]):
        detection_error = abs(cp - true_change_points[i])
        print("Detection error:", detection_error)
        assert (
            detection_error <= cp_detection_margin
        ), f"Detected change at index {cp}, expected close to {true_change_points[i]}"


def test_noise_sensitivity():
    """Test the sensitivity of both algorithms to different noise levels."""
    # Generate data with a single changepoint
    true_cps = [100]
    slopes = [0.1, -0.2]

    noise_levels = [0.01, 0.05, 0.1, 0.2, 0.3]
    max_deviations = []

    for noise_std in noise_levels:
        df = generate_continuous_piecewise_linear_signal(
            change_points=true_cps,
            slopes=slopes,
            n_samples=200,
            noise_std=noise_std,
            random_seed=42,
        )

        # Test Narrowest over threshold
        detector = SeededBinarySegmentation(
            ContinuousLinearTrendScore(), penalty=25, selection_method="narrowest"
        )

        cps = detector.fit_predict(df)

        if len(cps) == 1:
            deviation = abs(cps.iloc[0, 0] - true_cps[0])
            max_deviations.append(deviation)

    # Assert that for reasonable noise levels, we can detect the changepoint
    assert len(max_deviations) == len(
        noise_levels
    ), "Detection worked for low noise levels"
    assert (
        max(max_deviations) < 3
    ), f"Detection failed, with max deviation: {max(max_deviations)}"
    # For lower noise levels, the detection should be more accurate
    assert all(
        np.diff(max_deviations) >= 0
    ), "Detection accuracy doesn't improve with lower noise"


def test_multivariate_detection():
    """Test detection on multivariate continuous piecewise linear signals."""
    # Generate two different signals with the same changepoints
    change_points = [100, 200]
    slopes1 = [0.1, -0.2, 0.15]
    slopes2 = [0.05, 0.15, -0.1]

    df1 = generate_continuous_piecewise_linear_signal(
        change_points=change_points,
        slopes=slopes1,
        n_samples=300,
        noise_std=0.1,
        random_seed=42,
    )

    df2 = generate_continuous_piecewise_linear_signal(
        change_points=change_points,
        slopes=slopes2,
        n_samples=300,
        noise_std=0.1,
        random_seed=43,
    )

    # Combine into multivariate DataFrame
    df = pd.DataFrame({"signal1": df1["signal"], "signal2": df2["signal"]})

    # Test with SeededBinarySegmentation
    detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(),
        penalty=25,
        selection_method="greedy",
    )

    detected_cps = detector.fit_predict(df)

    # Assert we found the correct number of changepoints
    assert len(detected_cps) == len(
        change_points
    ), f"Expected {len(change_points)} changepoints, found {len(detected_cps)}"

    # Assert the changepoints are close to the true changepoints
    cp_detection_margin = 3
    for i, cp in enumerate(detected_cps["ilocs"]):
        assert (
            abs(cp - change_points[i]) <= cp_detection_margin
        ), f"Detected {cp}, expected close to {change_points[i]}"


def test_irregular_time_sampling():
    """Test ContinuousLinearTrendScore with irregular time sampling."""
    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.1, -0.2]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )
    df["sample_times"] = 50.0 + 2.0 * np.arange(n_samples)

    # Create irregular time sampling by selectively removing points
    # Keep all points around the changepoint for accurate detection
    np.random.seed(42)
    selection_mask = np.ones(n_samples, dtype=bool)

    # Remove ~30% of points away from the changepoint
    for region in [(0, 80), (90, 110), (120, 200)]:
        start, end = region
        region_len = end - start
        # Remove about 30% of points in each region
        to_remove = np.random.choice(
            np.arange(start, end), size=int(region_len * 0.3), replace=False
        )
        selection_mask[to_remove] = False

    # Apply the mask to create irregularly sampled data:
    irregular_df = df[selection_mask].copy()
    reverse_index_map = df.index[selection_mask]
    # Create a sample_times column that reflects the original indices
    true_change_point_time = df["sample_times"].iloc[true_change_points[0]]

    # Create detector with ContinuousLinearTrendScore with time_column
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(time_column="sample_times"),
        bandwidth=30,
        penalty=25,
    )

    # Fit and predict
    detected_cps = detector.fit_predict(irregular_df)

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Get the original index that corresponds to the detected changepoint
    detected_cp_idx = reverse_index_map[detected_cps.iloc[0, 0]]
    detected_cp_time = df["sample_times"].iloc[detected_cp_idx]

    # Assert the detected time is close to the true changepoint
    cp_detection_margin = 2  # Slightly larger margin for irregular sampling
    cp_time_detection_margin = 4.0
    assert (
        abs(detected_cp_idx - true_change_points[0]) <= cp_detection_margin
    ), f"Detection index {detected_cp_idx}, expected close to {true_change_points[0]}"

    # Assert the detected time is close to the true changepoint
    assert (
        abs(detected_cp_time - true_change_point_time) <= cp_time_detection_margin
    ), f"Detected time {detected_cp_time}, expected close \
        to {true_change_point_time}"

    # Test with SeededBinarySegmentation as well
    sbs_detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(time_column="sample_times"),
        selection_method="greedy",
        penalty=25,
    )

    # Fit and predict
    sbs_detected_cps = sbs_detector.fit_predict(irregular_df)

    # Assert we found 1 changepoint
    assert (
        len(sbs_detected_cps) == 1
    ), f"Expected 1 changepoint, found {len(sbs_detected_cps)}"

    # Get the original time that corresponds to the detected changepoint
    sbs_detected_cp_idx = reverse_index_map[sbs_detected_cps.iloc[0, 0]]
    sbs_detected_cp_time = df["sample_times"].iloc[sbs_detected_cp_idx]

    # Assert the detected time is close to the true changepoint
    assert (
        abs(sbs_detected_cp_idx - true_change_points[0]) <= cp_detection_margin
    ), f"SBS detected at time {sbs_detected_cp_time}, expected close \
          to {true_change_points[0]}"
    # Assert the detected time is close to the true changepoint
    assert (
        abs(sbs_detected_cp_time - true_change_point_time) <= cp_time_detection_margin
    ), f"SBS detected time {sbs_detected_cp_time}, expected close \
          to {true_change_point_time}"


def test_ignoring_irregular_time_sampling():
    """Test ContinuousLinearTrendScore when ignoring irregular time sampling."""
    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.1, -0.2]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )
    # df["sample_times"] = 50.0 + 2.0 * np.arange(n_samples)

    # Create irregular time sampling by selectively removing points
    # Keep all points around the changepoint for accurate detection
    np.random.seed(42)
    selection_mask = np.ones(n_samples, dtype=bool)

    # Remove ~30% of points away from the changepoint
    for region in [(0, 80), (90, 110), (120, 200)]:
        start, end = region
        region_len = end - start
        # Remove about 30% of points in each region
        to_remove = np.random.choice(
            np.arange(start, end), size=int(region_len * 0.4), replace=False
        )
        selection_mask[to_remove] = False

    # Apply the mask to create irregularly sampled data:
    irregular_df = df[selection_mask].copy()
    reverse_index_map = df.index[selection_mask]

    # Closest index to the true changepoint in the original data:
    true_cp_index = np.where(selection_mask[: true_change_points[0] + 1])[0][-1]

    # Create detector with ContinuousLinearTrendScore WITHOUT time_column
    mw_detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(),  # No time_column provided
        bandwidth=30,
        penalty=25,
    )

    # Fit and predict
    mw_detected_cps = mw_detector.fit_predict(irregular_df)

    # Assert we found 1 changepoint
    assert (
        len(mw_detected_cps) == 1
    ), f"Expected 1 changepoint, found {len(mw_detected_cps)}"

    # Get the detected changepoint index in the irregular data
    mw_detected_cp_idx = reverse_index_map[mw_detected_cps.iloc[0, 0]]

    # Without time information, the detected changepoint index should be different
    # from the one we would expect with time information
    lower_cp_detection_margin = 2
    upper_cp_detection_margin = 12
    assert (
        lower_cp_detection_margin
        < abs(mw_detected_cp_idx - true_cp_index)
        < upper_cp_detection_margin
    ), f"Detection could fail when ignoring irregular sampling: {mw_detected_cp_idx}"

    # Test with SeededBinarySegmentation as well
    sbs_detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(),  # No time_column provided
        selection_method="narrowest",
        penalty=25,
    )

    # Fit and predict
    sbs_detected_cps = sbs_detector.fit_predict(irregular_df)

    # Assert we found 1 changepoint
    assert (
        len(sbs_detected_cps) == 1
    ), f"Expected 1 changepoint, found {len(sbs_detected_cps)}"


def fit_indexed_linear_trend(xs: np.ndarray) -> tuple[float, float]:
    """Calculate the optimal linear trend for a given array.

    Assuming the time steps are [0, 1, 2, ..., n-1], we can optimize the calculation
    of the least squares intercept and slope.

    Parameters
    ----------
    xs : np.ndarray
        1D array of data points

    Returns
    -------
    tuple
        (slope, intercept) of the best-fit line
    """
    n_samples = len(xs)

    # For evenly spaced time steps [0, 1, 2, ..., n-1],
    # the mean time step is (n-1)/2.
    mean_t = (n_samples - 1) / 2.0

    # Optimized calculation for denominator:
    # sum of (t - mean_t)^2 = n*(n^2-1)/12
    denominator = n_samples * (n_samples * n_samples - 1) / 12.0

    # Calculate numerator: sum((t-mean_t)*(x-mean_x))
    # numerator = np.sum((np.arange(n) - mean_t) * (xs - mean_x))
    mean_x = np.mean(xs)
    numerator = 0.0
    for i in range(n_samples):
        numerator += (i - mean_t) * (xs[i] - mean_x)

    slope = numerator / denominator
    intercept = mean_x - slope * mean_t

    return intercept, slope


def linear_trend_score(
    starts: np.ndarray, splits: np.ndarray, ends: np.ndarray, X: np.ndarray
) -> np.ndarray:
    """Evaluate the continuous linear trend cost.

    Parameters
    ----------
    starts : np.ndarray
        Start indices of the first intervals (inclusive).
    splits : np.ndarray
        Split indices between the intervals (contained in second interval).
    ends : np.ndarray
        End indices of the second intervals (exclusive).
    X : np.ndarray
        Data to evaluate. Must be a 2D array.

    Returns
    -------
    costs : np.ndarray
        A 2D array of costs. One row for each interval. The number of columns
        is equal to the number of columns in the input data.
    """
    ### NOTE: Assume 'time' is index of the data. i.e. time = 0, 1, 2, ..., len(X)-1
    ###       This assumption could be changed later.
    n_intervals = len(starts)
    n_columns = X.shape[1]
    scores = np.zeros((n_intervals, n_columns))

    for i in range(n_intervals):
        start, split, end = starts[i], splits[i], ends[i]
        split_interval_trend_data = np.zeros((end - start, 3))
        split_interval_trend_data[:, 0] = 1.0  # Intercept
        # Whole interval slope:
        split_interval_trend_data[:, 1] = np.arange(end - start)  # Time steps

        # Change in slope from the split point:
        # trend data index starts at 0 from 'start'.
        # Continuous at the first point of the second interal, [split, end - 1]:
        split_interval_trend_data[(split - start) :, 2] = np.arange(end - split)

        # Calculate the slope and intercept for the whole interval:
        split_interval_linreg_res = np.linalg.lstsq(
            split_interval_trend_data, X[start:end, :]
        )
        split_interval_squared_residuals = split_interval_linreg_res[1]

        # By only regressing onto the first two columns, we can calculate the cost
        # without allowing for a change in slope at the split point.
        joint_interval_linreg_res = np.linalg.lstsq(
            split_interval_trend_data[:, [0, 1]], X[start:end, :]
        )
        joint_interval_squared_residuals = joint_interval_linreg_res[1]

        scores[i, :] = (
            joint_interval_squared_residuals - split_interval_squared_residuals
        )

    return scores


def test_linear_trend_score():
    # Straight linear trend test data
    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]], dtype=float)
    times = np.arange(1, X.shape[0] + 1)
    starts = np.array([0])
    splits = np.array([2])
    ends = np.array([5])

    # Expected output
    expected_scores = np.array([[0.0, 0.0]])

    # Call the function
    scores = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, times=times
    )

    # Check if the output matches the expected output
    assert np.allclose(
        scores, expected_scores
    ), f"Expected {expected_scores}, got {scores}"


def test_continuous_linear_trend_score_function():
    """Test that continuous_linear_trend_score function works as expected."""
    # Test data
    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]], dtype=float)
    starts = np.array([0])
    splits = np.array([2])
    ends = np.array([5])
    times = np.arange(X.shape[0])  # Default times [0, 1, 2, 3, 4]

    # Calculate scores using both methods
    lin_reg_score_values = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, times
    )
    analytical_score_values = analytical_cont_piecewise_linear_trend_score(
        starts, splits, ends, X
    )

    # Check if all methods give the same result
    assert np.allclose(lin_reg_score_values, analytical_score_values)


def test_linear_trend_score_with_different_times():
    """Test that linear_trend_score works with custom times."""
    # Test data
    np.random.seed(0)  # For reproducibility
    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]], dtype=float)
    X += np.random.normal(0, 0.1, X.shape)  # Add some noise
    starts = np.array([0])
    splits = np.array([2])
    ends = np.array([5])

    # Default times [0, 1, 2, 3, 4]
    default_times = np.arange(X.shape[0])
    # Custom evenly spaced times
    custom_times = np.array([0, 2, 4, 6, 8])

    # Calculate scores
    scores_default = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, default_times
    )
    scores_custom = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, custom_times
    )

    # Since the times are evenly spaced but with different scales,
    # the scores should be the same
    assert np.allclose(scores_default, scores_custom)


def test_multiple_segments():
    """Test with multiple segments to evaluate."""
    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8]], dtype=float)
    # Add some noise:
    # np.random.seed(0)  # For reproducibility
    X += np.random.normal(0, 1.0, X.shape)  # Add some noise
    starts = np.array([0, 2])
    splits = np.array([2, 5])
    ends = np.array([5, 7])
    times = np.arange(X.shape[0])

    # Calculate scores using both methods
    scores_func = analytical_cont_piecewise_linear_trend_score(starts, splits, ends, X)
    scores_with_times = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, times
    )

    # Check if both methods give the same result
    assert np.allclose(scores_func, scores_with_times)


def test_datetime_indices():
    """Test ContinuousLinearTrendScore with datetime indices."""

    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.5, -0.5]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )

    # Create datetime index starting from today
    start_date = datetime.datetime.now()
    datetime_index = [start_date + datetime.timedelta(days=i) for i in range(n_samples)]
    df["datetime"] = datetime_index

    ## Sane way of creating a numeric time column:
    df["time_elapsed"] = df["datetime"] - df.loc[0, "datetime"]
    df["num_timestamp"] = (
        df["time_elapsed"].dt.total_seconds() / 3600.0
    )  # Convert to hours

    # Create detector with ContinuousLinearTrendScore with datetime column
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(time_column="num_timestamp"),
        bandwidth=30,
        penalty=25,
    )

    # Fit and predict:
    detected_cps = detector.fit_predict(df.drop(columns=["datetime", "time_elapsed"]))

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Assert the changepoint is close to the true changepoint
    detected_cp = detected_cps.iloc[0, 0]
    cp_detection_margin = 3
    assert (
        abs(detected_cp - true_change_points[0]) <= cp_detection_margin
    ), f"Detected {detected_cp}, expected close to {true_change_points[0]}"


def test_numpy_datetime64():
    """Test ContinuousLinearTrendScore with numpy datetime64 type."""
    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.1, -0.2]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )

    # Create numpy datetime64 column
    start_date = np.datetime64("2023-01-01")
    dt_array = np.array([start_date + np.timedelta64(i, "h") for i in range(n_samples)])
    df["np_datetime"] = dt_array

    # Create a numeric timestamp:
    df["num_timestamp"] = (
        df["np_datetime"] - df["np_datetime"][0]
    ).dt.total_seconds() / (60.0 * 60.0)  # Convert to hours

    # Create detector with ContinuousLinearTrendScore with datetime column:
    detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(time_column="num_timestamp"),
        penalty=25,
        selection_method="narrowest",
    )

    # Fit and predict
    detected_cps = detector.fit_predict(df.drop(columns=["np_datetime"]))

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Assert the changepoint is close to the true changepoint
    detected_cp = detected_cps.iloc[0, 0]
    cp_detection_margin = 3
    assert (
        abs(detected_cp - true_change_points[0]) <= cp_detection_margin
    ), f"Detected {detected_cp}, expected close to {true_change_points[0]}"


def test_pandas_timestamp():
    """Test ContinuousLinearTrendScore with pandas Timestamp type."""
    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.1, -0.2]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )

    # Create pandas Timestamp column
    start_date = pd.Timestamp("2023-01-01")
    pd_timestamps = [start_date + pd.Timedelta(days=i) for i in range(n_samples)]
    df["pd_timestamp"] = pd_timestamps

    # Create detector with ContinuousLinearTrendScore with datetime column
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(time_column="pd_timestamp"),
        bandwidth=30,
        penalty=25,
    )

    with pytest.raises(
        TypeError,
        match=re.escape(
            "float() argument must be a string or a real number, not 'Timestamp'"
        ),
    ):
        # Fit and predict should raise an error:
        _ = detector.fit_predict(df)


def test_date_objects():
    """Test ContinuousLinearTrendScore with Python date objects."""

    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.5, -0.5]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )

    # Create Python date objects
    start_date = datetime.date(2023, 1, 1)
    dates = [start_date + datetime.timedelta(days=i) for i in range(n_samples)]
    # Call 'convert_dtypes' to make sure the date column is of 'dtype' "date":
    df["date"] = pd.to_datetime(pd.Series(dates))
    # Create a numeric timestamp:
    df["days_since_first"] = (
        df["date"] - df.loc[0, "date"]
    ).dt.total_seconds() / pd.Timedelta(days=1).total_seconds()  # Convert to hours

    # Create detector with ContinuousLinearTrendScore with date column
    detector = SeededBinarySegmentation(
        change_score=ContinuousLinearTrendScore(time_column="days_since_first"),
        penalty=25,
        selection_method="narrowest",
    )

    # Fit and predict
    detected_cps = detector.fit_predict(df.drop(columns=["date"]))

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Assert the changepoint is close to the true changepoint
    detected_cp = detected_cps.iloc[0, 0]
    cp_detection_margin = 3
    assert (
        abs(detected_cp - true_change_points[0]) <= cp_detection_margin
    ), f"Detected {detected_cp}, expected close to {true_change_points[0]}"


def test_non_uniform_datetime_sampling():
    """Test ContinuousLinearTrendScore with non-uniform datetime sampling."""

    # Generate data with a single changepoint
    true_change_points = [100]
    slopes = [0.5, -0.5]
    n_samples = 200

    df = generate_continuous_piecewise_linear_signal(
        change_points=true_change_points,
        slopes=slopes,
        n_samples=n_samples,
        noise_std=1.0,
        random_seed=42,
    )

    # Create non-uniform datetime sampling
    start_date = datetime.datetime(2023, 1, 1)
    np.random.seed(42)
    # Add random hours (0-23) to each day
    random_hours = np.random.randint(0, 24, n_samples)
    datetimes = [
        start_date + datetime.timedelta(days=int(i), hours=int(h))
        for i, h in enumerate(random_hours)
    ]
    df["datetime"] = pd.to_datetime(pd.Series(datetimes))
    df["num_timestamp"] = (
        df["datetime"] - df.loc[0, "datetime"]
    ).dt.total_seconds() / pd.Timedelta(days=1).total_seconds()
    # Note the timestamp at the true changepoint for later comparison
    true_cp_time = df["datetime"].iloc[true_change_points[0]]

    # Create detector with ContinuousLinearTrendScore with datetime column
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(time_column="num_timestamp"),
        bandwidth=30,
        penalty=25,
    )

    # Fit and predict
    detected_cps = detector.fit_predict(df.drop(columns=["datetime"]))

    # Assert we found 1 changepoint
    assert len(detected_cps) == 1, f"Expected 1 changepoint, found {len(detected_cps)}"

    # Assert the changepoint is close to the true changepoint
    detected_cp = detected_cps.iloc[0, 0]
    cp_detection_margin = 3
    assert (
        abs(detected_cp - true_change_points[0]) <= cp_detection_margin
    ), f"Detected {detected_cp}, expected close to {true_change_points[0]}"

    # Also check that the datetime is close
    detected_time = df["datetime"].iloc[detected_cp]
    time_diff = abs((detected_time - true_cp_time).total_seconds())
    max_allowed_diff = pd.Timedelta(days=4).total_seconds()
    assert (
        time_diff <= max_allowed_diff
    ), f"Detected time {detected_time} too far from true time {true_cp_time}"


def test_large_timestamp_values():
    """Test ContinuousLinearTrendScore with extremely large timestamp values."""
    # Generate test data
    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8]], dtype=float)
    starts = np.array([0])
    splits = np.array([2])
    ends = np.array([5])

    # Create extremely large timestamps (e.g., microseconds since epoch)
    large_times = 1.676e13 + np.arange(X.shape[0]) * 1e9  # Large timestamp values

    # Calculate scores using the linear regression method
    scores = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, large_times
    )

    # Check if scores contain NaN values
    assert np.any(np.isnan(scores)), "Expected NaN scores with large timestamp values"

    # Create a detector with a time column
    df = pd.DataFrame({"signal1": X[:, 0], "signal2": X[:, 1], "time": large_times})

    # Test graceful handling in MovingWindow
    detector = MovingWindow(
        change_score=ContinuousLinearTrendScore(time_column="time"),
        bandwidth=3,
        penalty=1.0,  # Low penalty to encourage changepoint detection
    )

    # The detector should not crash, even with extreme timestamp values
    try:
        detected_cps = detector.fit_predict(df)
        # If we get here, it didn't crash
        assert True, "MovingWindow handled large timestamps without crashing"
    except np.linalg.LinAlgError:
        # Expected behavior - singular matrix in least squares
        assert True, "Expected LinAlgError with large timestamps"
    except Exception as e:
        pytest.fail(f"Unexpected error type: {type(e)}, message: {str(e)}")

    # Alternative approach: scale down the timestamps
    normalized_times = (large_times - large_times[0]) / 1e9

    scores_normalized = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, normalized_times
    )

    # Normalized timestamps should work properly and not produce NaN
    assert not np.any(
        np.isnan(scores_normalized)
    ), "Normalized timestamps should not produce NaN"

    # Check that the detector works with normalized timestamps
    df["normalized_time"] = normalized_times
    detector_normalized = MovingWindow(
        change_score=ContinuousLinearTrendScore(time_column="normalized_time"),
        bandwidth=3,
        penalty=1.0,
    )

    # This should now work without numerical issues
    detected_cps = detector_normalized.fit_predict(df)
    assert isinstance(
        detected_cps, pd.DataFrame
    ), "Should return valid changepoints with normalized timestamps"


def test_three_point_segment():
    """Test with a three-point segment."""
    X = np.array([[1, 2], [2, 3], [3, 4]], dtype=float)
    # Add some noise:
    np.random.seed(0)  # For reproducibility
    X += np.random.normal(0, 1.0, X.shape)  # Add some noise
    starts = np.array([0])
    splits = np.array([1])
    ends = np.array([3])
    times = np.arange(X.shape[0])

    # Calculate scores using both methods
    scores_func = analytical_cont_piecewise_linear_trend_score(starts, splits, ends, X)
    scores_with_times = lin_reg_cont_piecewise_linear_trend_score(
        starts, splits, ends, X, times
    )
    assert np.allclose(
        scores_func, scores_with_times
    ), f"Expected {scores_func}, got {scores_with_times}"



"""Tests for continous piecewise linear data generation."""

import pandas as pd
import pytest

from skchange.datasets import generate_continuous_piecewise_linear_data


def test_generate_continuous_piecewise_linear_data_default():
    df = generate_continuous_piecewise_linear_data()
    assert isinstance(df, pd.DataFrame)
    assert not df.empty


@pytest.mark.parametrize(
    "slopes",
    [
        None,
        1,
        [1],
        [2, 3],
        [0.5, -0.5, 1],
    ],
)
def test_generate_continuous_piecewise_linear_data_valid_slopes(
    slopes: float | list[float],
):
    """Test that invalid slopes raise ValueError."""
    df = generate_continuous_piecewise_linear_data(slopes=slopes)
    assert not df.empty


def test_generate_continuous_piecewise_linear_data_invalid_noise_std():
    """Test that invalid noise_std raises ValueError."""
    with pytest.raises(ValueError):
        generate_continuous_piecewise_linear_data(noise_std=-1)



import numpy as np
import pytest

from skchange.costs import L2Cost


def test_l2_cost_check_fixed_param():
    cost = L2Cost()
    X = np.array([[1, 2], [3, 4]])
    mean = cost._check_fixed_param(1.0, X)
    np.testing.assert_array_equal(mean, np.array([1.0]))

    mean = cost._check_fixed_param([1.0, 2.0], X)
    np.testing.assert_array_equal(mean, np.array([1.0, 2.0]))

    with pytest.raises(ValueError):
        cost._check_fixed_param([1.0, 2.0, 3.0], X)



"""The Moving Window algorithm for multiple changepoint detection."""

__author__ = ["Tveten"]
__all__ = ["MovingWindow"]

import numpy as np
import pandas as pd

from ..base import BaseIntervalScorer
from ..change_scores import CUSUM, to_change_score
from ..compose.penalised_score import PenalisedScore
from ..utils.numba import njit
from ..utils.numba.general import where
from ..utils.validation.data import check_data
from ..utils.validation.interval_scorer import check_interval_scorer
from ..utils.validation.parameters import check_in_interval, check_larger_than
from ..utils.validation.penalties import check_penalty
from .base import BaseChangeDetector


@njit
def make_extended_moving_window_cuts(
    n_samples: int,
    bandwidth: int,
    min_size: int,
) -> np.ndarray:
    splits = np.arange(min_size, n_samples - min_size + 1)

    starts = splits - bandwidth
    starts[starts < 0] = 0
    max_start = n_samples - 2 * bandwidth
    starts[starts > max_start] = max_start

    ends = splits + bandwidth
    ends[ends > n_samples] = n_samples
    min_end = 2 * bandwidth
    ends[ends < min_end] = min_end

    cuts = np.column_stack((starts, splits, ends))
    return cuts


def transform_moving_window(
    penalised_score: BaseIntervalScorer,
    bandwidth: int,
) -> np.ndarray:
    penalised_score.check_is_penalised()
    penalised_score.check_is_fitted()

    n_samples = penalised_score.n_samples
    cuts = make_extended_moving_window_cuts(
        n_samples, bandwidth, penalised_score.min_size
    )
    scores = np.repeat(np.nan, n_samples)
    scores[cuts[:, 1]] = penalised_score.evaluate(cuts).reshape(-1)
    return scores


def transform_multiple_moving_window(
    penalised_score: BaseIntervalScorer,
    bandwidths: np.ndarray,
) -> np.ndarray:
    n_samples = penalised_score.n_samples
    scores = np.full((n_samples, len(bandwidths)), np.nan)
    for i, bw in enumerate(bandwidths):
        scores[:, i] = transform_moving_window(penalised_score, bw)
    return scores


@njit
def get_candidate_changepoints(
    scores: np.ndarray,
) -> tuple[list[int], list[tuple[int, int]]]:
    detection_intervals = where(scores > 0)
    changepoints = []
    for start, end in detection_intervals:
        cpt = start + np.argmax(scores[start:end])
        changepoints.append(cpt)
    return changepoints, detection_intervals


@njit
def select_changepoints_by_detection_length(
    scores: np.ndarray, min_detection_interval: int
) -> list:
    candidate_cpts, detection_intervals = get_candidate_changepoints(scores)
    cpts = [
        cpt
        for cpt, interval in zip(candidate_cpts, detection_intervals)
        if interval[1] - interval[0] >= min_detection_interval
    ]

    return cpts


@njit
def select_changepoints_by_local_optimum(
    scores: np.ndarray, selection_bandwidth: int
) -> list:
    candidate_cpts, _ = get_candidate_changepoints(scores)
    cpts = [
        cpt
        for cpt in candidate_cpts
        if np.isclose(
            scores[cpt],
            np.max(
                scores[
                    max(cpt - selection_bandwidth, 0) : cpt + selection_bandwidth + 1
                ]
            ),
        )
    ]

    return cpts


@njit
def select_changepoints_by_bottom_up(
    scores: np.ndarray, bandwidths: np.ndarray, local_optimum_fraction: float
) -> list:
    bandwidths = sorted(bandwidths)
    candidate_cpts = []
    for i, bw in enumerate(bandwidths):
        local_optimum_bandwidth = int(local_optimum_fraction * bw)
        candidate_cpts_bw = select_changepoints_by_local_optimum(
            scores[:, i], local_optimum_bandwidth
        )
        for candidate_cpt in candidate_cpts_bw:
            candidate_cpts.append((candidate_cpt, bw))

    cpts = [candidate_cpts[0][0]]
    for candidate_cpt, bw in candidate_cpts[1:]:
        distance_to_closest = np.min(np.abs(candidate_cpt - np.array(cpts)))
        local_optimum_bandwidth = int(local_optimum_fraction * bw)
        if distance_to_closest >= local_optimum_bandwidth:
            cpts.append(candidate_cpt)

    return cpts


class MovingWindow(BaseChangeDetector):
    """Moving window algorithm for multiple change-point detection.

    The MOSUM (moving sum) algorithm [1]_, but generalized to allow for any penalised
    and unpenalised change score. The basic algorithm runs a test statistic for a
    single change-point across the data in a moving window fashion.
    In each window, the data is split into two equal halves with `bandwidth` samples
    on either side of a split point.
    This process generates a time series of penalised scores, which are used to generate
    candidate change-points as local maxima within intervals where the penalised scores
    are all above zero.
    The final set of change-points is selected from the candidate change-points using
    one of the two selection methods described in [2]_.

    Several of the extensions available in the mosum R package [2]_ are also available
    in this implementation, including the ability to use multiple bandwidths. The
    CUSUM-type boundary extension for computing the test statistic for candidate change-
    points less than `bandwidth` samples from the start and end of the data is also
    implemented by default.

    Parameters
    ----------
    change_score : BaseIntervalScorer, optional, default=CUSUM()
        The change score to use in the algorithm. If a cost is given, it is
        converted to a change score using the `ChangeScore` class.
    penalty : np.ndarray or float, optional, default=None
        The penalty to use for change detection. If the penalty is
        penalised (`change_score.get_tag("is_penalised")`) the penalty will
        be ignored. The different types of penalties are as follows:

        * ``float``: A constant penalty applied to the sum of scores across all
          variables in the data.
        * ``np.ndarray``: A penalty array of the same length as the number of
          columns in the data, where element ``i`` of the array is the penalty for
          ``i+1`` variables being affected by a change. The penalty array
          must be positive and increasing (not strictly). A penalised score with a
          linear penalty array is faster to evaluate than a nonlinear penalty array.
        * ``None``: A default penalty is created in `predict` based on the fitted
          score using the `make_bic_penalty` function.

    bandwidth : int or list of int, default=20
        The bandwidth is the number of samples on either side of a candidate
        change-point. Must be 1 or greater. If a list of bandwidths is given, the
        algorithm will run for each bandwidth in the list and combine the results
        accoring to the "bottom-up" merging approach described in [2]_. A fibonacci
        sequence of bandwidths is recommended for multiple bandwidths by the authors
        in [2]_.
    selection_method : str, default="local_optimum"
        The method used to select the final set of change-points from a set of candidate
        change-points. The options are:

        * ``"detection_length"``: Accepts a candidate change-point if the
          ``min_detection_fraction * bandwidth`` consecutive penalised scores are above
          zero. Corresponds to the epsilon-criterion in [2]_. This method is only
          available for a single bandwidth.
        * ``"local_optimum"``: Accepts a candidate change-point if it is the local
          maximum in the scores within a neighbourhood of size
          ``local_optimum_fraction * bandwidth``. Corresponds to the eta-criterion
          in [2]_. This method is used within the "bottom-up" merging approach if
          multiple bandwidths are given.

    min_detection_fraction : float, default=0.2
        The minimum size of the detection interval for a candidate change-point to be
        accepted in the ``"detection_length"`` selection method.
        be between ``0`` (exclusive) and ``1/2`` (exclusive).
    local_optimum_fraction : float, default=0.4
        The size of the neighbourhood around a candidate change-point used in the
        ``"local_optimum"`` selection method. Must be larger than or equal to ``0``.

    References
    ----------
    .. [1] Eichinger, B., & Kirch, C. (2018). A MOSUM procedure for the estimation of
       multiple random change points.

    .. [2] Meier, A., Kirch, C., & Cho, H. (2021). mosum: A package for moving sums in
       change-point analysis. Journal of Statistical Software, 97, 1-42.

    Examples
    --------
    >>> from skchange.change_detectors import MovingWindow
    >>> from skchange.datasets import generate_alternating_data
    >>> df = generate_alternating_data(n_segments=4, mean=10, segment_length=100, p=5)
    >>> detector = MovingWindow()
    >>> detector.fit_predict(df)
       ilocs
    0    100
    1    200
    2    300
    """

    _tags = {
        "authors": ["Tveten"],
        "maintainers": ["Tveten"],
        "fit_is_empty": True,
    }

    def __init__(
        self,
        change_score: BaseIntervalScorer | None = None,
        penalty: np.ndarray | float | None = None,
        bandwidth: int | list = 20,
        selection_method: str = "local_optimum",
        min_detection_fraction: float = 0.2,
        local_optimum_fraction: float = 0.4,
    ):
        self.change_score = change_score
        self.penalty = penalty
        self.bandwidth = bandwidth
        self.selection_method = selection_method
        self.min_detection_fraction = min_detection_fraction
        self.local_optimum_fraction = local_optimum_fraction
        super().__init__()

        _score = CUSUM() if change_score is None else change_score
        check_interval_scorer(
            _score,
            "change_score",
            "MovingWindow",
            required_tasks=["cost", "change_score"],
            allow_penalised=True,
        )
        _change_score = to_change_score(_score)

        check_penalty(penalty, "penalty", "MovingWindow")
        self._penalised_score = (
            _change_score.clone()  # need to avoid modifying the input change_score
            if _change_score.get_tag("is_penalised")
            else PenalisedScore(_change_score, penalty)
        )

        if isinstance(bandwidth, int):
            check_larger_than(1, bandwidth, "bandwidth")
            self._bandwidth = np.array([bandwidth])
        elif isinstance(bandwidth, list):
            if len(bandwidth) == 0:
                raise ValueError("`bandwidth` must be a non-empty list.")
            if not all(isinstance(bw, int) for bw in bandwidth):
                raise TypeError("All elements of `bandwidth` must be integers.")
            if any(bw < 1 for bw in bandwidth):
                raise ValueError("All elements of `bandwidth` must be greater than 0.")
            self._bandwidth = np.array(bandwidth)
        else:
            raise TypeError(
                "`bandwidth` must be an integer or a list of integers. "
                f"Got {type(bandwidth)}."
            )

        check_in_interval(
            pd.Interval(0, 1 / 2, closed="neither"),
            min_detection_fraction,
            "min_detection_fraction",
        )
        check_larger_than(0, local_optimum_fraction, "local_optimum_fraction")

        valid_selection_methods = ["local_optimum", "detection_length"]
        if selection_method not in valid_selection_methods:
            raise ValueError(
                f"`selection_method` must be one of {valid_selection_methods}."
                f" Got {selection_method}."
            )
        if len(self._bandwidth) > 1 and self.selection_method == "detection_length":
            raise ValueError(
                "The selection method `detection_length` is not supported for multiple"
                " bandwidths. Use `'local_optimum'` instead."
            )

        self.clone_tags(_change_score, ["distribution_type"])

    def _transform_scores(self, X: pd.DataFrame | pd.Series) -> pd.DataFrame:
        """Return scores for predicted labels on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame, pd.Series or np.ndarray
            Data to score (time series).

        Returns
        -------
        scores : pd.DataFrame with same index as X
            Scores for sequence `X`.

        Attributes
        ----------
        fitted_score : BaseIntervalScorer
            The fitted penalised change score used for the detection.
        """
        X = check_data(
            X,
            min_length=2 * max(self._bandwidth),
            min_length_name="2*max(bandwidth)",
        )

        self.fitted_score: BaseIntervalScorer = self._penalised_score.clone()
        self.fitted_score.fit(X)
        scores = transform_multiple_moving_window(self.fitted_score, self._bandwidth)
        formatted_scores = pd.DataFrame(
            scores,
            index=X.index,
            columns=pd.Index([bw for bw in self._bandwidth], name="bandwidth"),
        )
        return formatted_scores

    def _predict(self, X: pd.DataFrame | pd.Series) -> pd.DataFrame:
        """Detect events in test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame
            Time series to detect change points in.

        Returns
        -------
        y_sparse : pd.DataFrame
            A `pd.DataFrame` with a range index and one column:
            * ``"ilocs"`` - integer locations of the changepoints.

        Attributes
        ----------
        fitted_score : BaseIntervalScorer
            The fitted penalised change score used for the detection.
        scores : pd.Series
            The detection scores obtained by the `transform_scores` method.
        """
        self.scores: pd.DataFrame = self.transform_scores(X)

        if self.selection_method == "detection_length":
            min_detection_length = int(self.min_detection_fraction * self.bandwidth)
            changepoints = select_changepoints_by_detection_length(
                self.scores.values.reshape(-1), min_detection_length
            )
        elif self.selection_method == "local_optimum" and len(self._bandwidth) == 1:
            local_optimum_bandwidth = int(self.local_optimum_fraction * self.bandwidth)
            changepoints = select_changepoints_by_local_optimum(
                self.scores.values.reshape(-1), local_optimum_bandwidth
            )
        else:
            changepoints = select_changepoints_by_bottom_up(
                self.scores.values, self._bandwidth, self.local_optimum_fraction
            )

        return self._format_sparse_output(changepoints)

    @classmethod
    def get_test_params(cls, parameter_set: str = "default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return ``"default"`` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skchange.costs import GaussianCost, L2Cost

        params = [
            {"change_score": L2Cost(), "bandwidth": 5, "penalty": 20},
            {"change_score": GaussianCost(), "bandwidth": 5, "penalty": 30},
        ]
        return params



import numpy as np
import scipy.special as sp
import scipy.stats as st

from skchange.utils.numba.stats import digamma, kurtosis, log_gamma, trigamma


def test_numba_kurtosis():
    seed = 523
    n_samples = 100
    p = 5
    t_dof = 5.0

    np.random.seed(seed)

    # Spd covariance matrix:
    random_nudge = np.random.randn(p).reshape(-1, 1)
    cov = np.eye(p) + 0.5 * random_nudge @ random_nudge.T

    mean = np.arange(p) * (-1 * np.ones(p)).cumprod()

    mv_t_dist = st.multivariate_t(loc=mean, shape=cov, df=t_dof)
    mv_t_samples = mv_t_dist.rvs(n_samples)

    sample_medians = np.median(mv_t_samples, axis=0)
    centered_samples = mv_t_samples - sample_medians

    # Test numba kurtosis:
    numba_kurtosis_val = kurtosis(centered_samples)
    scipy_kurtosis_val = st.kurtosis(centered_samples, fisher=True, bias=True, axis=0)

    assert np.all(np.isfinite(numba_kurtosis_val)), "Numba kurtosis should be finite."
    assert np.all(np.isfinite(scipy_kurtosis_val)), "Scipy kurtosis should be finite."
    (
        np.testing.assert_allclose(numba_kurtosis_val, scipy_kurtosis_val),
        "Kurtosis is off.",
    )


def test_numba_digamma():
    # Simpler test for digamma:
    np.random.seed(5125)
    random_vals = np.random.rand(100) * 10.0 + 0.5
    numba_digamma_vals = np.array([digamma(val) for val in random_vals])
    scipy_digamma_vals = sp.digamma(random_vals)
    np.testing.assert_allclose(numba_digamma_vals, scipy_digamma_vals, atol=1e-6)


def test_numba_digamma_on_small_values_gives_nan():
    # Test that digamma is not evaluated for small values:
    np.random.seed(5125)
    random_vals = np.random.rand(100) * 0.5
    numba_digamma_vals = np.array([digamma(val) for val in random_vals])
    less_than_threshold = random_vals <= 1.0e-2

    assert less_than_threshold.sum() > 0, "No small values found for digamma test."
    assert np.isnan(
        numba_digamma_vals[less_than_threshold]
    ).all(), "Numba digamma should be NaN on small values (<= 1.0e-2)."
    assert np.isfinite(
        numba_digamma_vals[~less_than_threshold]
    ).all(), "Numba digamma should be finite on values > 1.0e-2."


def test_numba_trigamma():
    np.random.seed(3125)
    random_vals = np.random.rand(100) * 10.0 + 0.5
    numba_trigamma_vals = np.array([trigamma(val) for val in random_vals])
    scipy_trigamma_vals = sp.polygamma(1, random_vals)
    np.testing.assert_allclose(numba_trigamma_vals, scipy_trigamma_vals, atol=1e-4)


def test_numba_trigamma_on_small_values_gives_nan():
    # Test that trigamma is not evaluated for small values:
    np.random.seed(5125)
    random_vals = np.random.rand(100) * 0.5
    numba_trigamma_vals = np.array([trigamma(val) for val in random_vals])
    less_than_threshold = random_vals <= 1.0e-2

    assert less_than_threshold.sum() > 0, "No small values found for trigamma test."
    assert np.isnan(
        numba_trigamma_vals[less_than_threshold]
    ).all(), "Numba trigamma should be NaN on small values (<= 1.0e-2)."
    assert np.isfinite(
        numba_trigamma_vals[~less_than_threshold]
    ).all(), "Numba trigamma should be finite on values > 1.0e-2."


def test_numba_log_gamma():
    np.random.seed(3125)
    # Our numba implementation should not be evaluated below 1.0:
    random_vals = np.random.rand(100) * 10.0 + 1.0
    numba_log_gamma_vals = np.array([log_gamma(val) for val in random_vals])
    scipy_log_gamma_vals = sp.gammaln(random_vals)
    np.testing.assert_allclose(numba_log_gamma_vals, scipy_log_gamma_vals, atol=1.0e-3)


def test_numba_log_gamma_on_small_values_gives_nan():
    # Test that log_gamma is not evaluated for small values:
    np.random.seed(5125)
    random_vals = np.random.rand(100) * 0.5
    numba_log_gamma_vals = np.array([log_gamma(val) for val in random_vals])
    less_than_threshold = random_vals <= 1.0e-2

    assert less_than_threshold.sum() > 0, "No small values found for log_gamma test."
    assert np.isnan(
        numba_log_gamma_vals[less_than_threshold]
    ).all(), "Numba log_gamma should be NaN on small values (<= 1.0e-2)."
    assert np.isfinite(
        numba_log_gamma_vals[~less_than_threshold]
    ).all(), "Numba log_gamma should be finite on values > 1.0e-2."



"""Interactive exploration of Seeded Binary Segmentation."""

import plotly.express as px

from skchange.change_detectors import SeededBinarySegmentation
from skchange.datasets import generate_alternating_data
from skchange.utils.benchmarking.profiler import Profiler

df = generate_alternating_data(
    n_segments=2, mean=10, segment_length=20, p=1, random_state=7
)
detector = SeededBinarySegmentation(growth_factor=2)
detector.fit_predict(df)

px.line(df)
px.scatter(detector.scores, x="argmax_cpt", y="score", hover_data=["start", "end"])


# Profiling
n = int(1e6)
df = generate_alternating_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = SeededBinarySegmentation(growth_factor=1.5, min_segment_length=10)
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()


# Test tuning
df_train = generate_alternating_data(
    n_segments=1, mean=0, segment_length=10000, p=1, random_state=9
)
df_test = generate_alternating_data(
    n_segments=10, mean=5, segment_length=1000, p=1, random_state=5
)
detector = SeededBinarySegmentation(threshold_scale=None, min_segment_length=10)
detector.fit(df_train)
changepoints = detector.predict(df_test)



"""Functions for loading datasets that ships with `skchange`."""

import os

import pandas as pd


def load_hvac_system_data():
    """Load the heating, ventilation and air conditioning (HVAC) system dataset.

    The dataset contains time series of vibration magnitude measurements from two
    different HVAC systems. 30 days of sensor measurements are available for each unit,
    with a sampling rate of 10 minutes.

    The aim of analysing the data is to detect when each hvac system normally
    turns on and off, such that anomalies from their regular schedule can be detected.
    True labels are not available, but it is fairly easy to identify the normal
    schedule from the data and observe when the units deviate from it.

    The data has been provided by the company Soundsensing:
    https://www.soundsensing.no/.

    Returns
    -------
    pd.DataFrame
        The HVAC system dataset. It has a multi-index with two levels:

        1. "unit_id": A string identifier for each hvac unit.
        2. "time": A datetime index with the time of each measurement.

        There's one column:

        1. "vibration": A float column with the vibration magnitude measurements.
    """
    this_file_dir = os.path.dirname(__file__)
    file_path = os.path.join(this_file_dir, "data", "hvac_system", "data.csv")
    df = pd.read_csv(file_path).iloc[:, 1:]
    df["time"] = pd.to_datetime(df["time"])
    df = df.set_index(["unit_id", "time"])
    return df



#!/usr/bin/env python3 -u
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
# TODO: Adapt this script to skchange when it's time for its first release.

"""Do-nothing script for making a release.

This idea comes from here:
- https://blog.danslimmon.com/2019/07/15/do-nothing-scripting-the-key-to
-gradual-automation/

The script is adapted from:
- https://github.com/sktime/sktime/blob/main/build_tools/make_release.py
"""

__author__ = ["mloning"]

import codecs
import os
import re
import webbrowser

import colorama

ROOT_DIR = os.path.abspath(os.path.dirname(__file__)).replace("build_tools", "")
PACKAGE_NAME = "skchange"


class URLs:
    """Container class for URLs."""

    DOCS_LOCAL = "file://" + os.path.realpath(
        os.path.join(ROOT_DIR, "docs/build/html/index.html")
    )
    DOCS_ONLINE = "https://skchange.readthedocs.io/"
    PYPI = f"https://pypi.org/simple/{PACKAGE_NAME}/"


def read(*parts):
    """Read from parts."""
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(ROOT_DIR, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    """Find version string."""
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


def open_website(website):
    """Open website."""
    webbrowser.open(website, new=1)


def colored(msg, color=None, style=None):
    """Print message with colors."""
    colors = {
        "red": colorama.Fore.RED,
        "green": colorama.Fore.GREEN,
        "cyan": colorama.Fore.CYAN,
        "yellow": colorama.Fore.YELLOW,
        "magenta": colorama.Fore.MAGENTA,
        None: "",
    }
    styles = {
        "bright": colorama.Style.BRIGHT,
        "dim": colorama.Style.DIM,
        None: "",
    }
    pre = colors[color] + styles[style]
    post = colorama.Style.RESET_ALL
    return f"{pre}{msg}{post}"


def cprint(msg, color=None, style=None):
    """Coloured printing."""
    print(colored(msg, color=color, style=style))  # noqa


def wait_for_enter():
    """Wait for Enter."""
    input(colored("\nPress Enter to continue", style="dim"))
    print()  # noqa


class Step:
    """Abstraction for release step."""

    def pre(self, context):
        """Pre-step."""
        pass

    def post(self, context):
        """Post-step."""
        wait_for_enter()

    def run(self, context):
        """Run step."""
        try:
            self.pre(context)
            self.action(context)
            self.post(context)
        except KeyboardInterrupt:
            cprint("\nInterrupted.", color="red")
            raise SystemExit(1)

    @staticmethod
    def instruct(msg):
        """Instruction."""
        cprint(msg, color="green")

    def print_run(self, msg):
        """Print run step."""
        cprint("Run:", color="cyan", style="bright")
        self.print_cmd(msg)

    @staticmethod
    def print_cmd(msg):
        """Print cmd step."""
        cprint("\t" + msg, color="cyan", style="bright")

    @staticmethod
    def do_cmd(cmd):
        """Wait for confirmation."""
        cprint(f"Going to run: {cmd}", color="magenta", style="bright")
        wait_for_enter()
        os.system(cmd)

    def action(self, context):
        """Carry out action."""
        raise NotImplementedError("abstract method")


class ConfirmGitStatus(Step):
    """Confirm git status."""

    def __init__(self, branch):
        self.branch = branch

    def action(self, context):
        """Carry out action."""
        self.instruct(
            f"Make sure you're on: {self.branch}, your local "
            f"branch is up-to-date, and all new changes are merged "
            f"in."
        )
        self.do_cmd(f"git checkout {self.branch}")
        self.do_cmd("git pull")


class UpdateChangelog(Step):
    """Update the changelog."""

    def action(self, context):
        """Carry out action."""
        self.instruct(f"Update CHANGELOG for version: {context['version']}")


class UpdateReadme(Step):
    """Update the readme."""

    def action(self, context):
        """Carry out action."""
        self.instruct(f"Update README for version: {context['version']}")


class UpdateVersion(Step):
    """Update skchange version."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Update pyproject.toml and skchange/__init__.py with new version")

    def post(self, context):
        """Post-action step."""
        wait_for_enter()
        context["version"] = find_version(context["package_name"], "__init__.py")


class MakeClean(Step):
    """Make clean."""

    def action(self, context):
        """Carry out action."""
        self.do_cmd("make clean")


class MakeDocs(Step):
    """Make docs."""

    def action(self, context):
        """Carry out action."""
        self.do_cmd("make docs")


class MakeDist(Step):
    """Make dist."""

    def action(self, context):
        """Carry out action."""
        self.do_cmd("make dist")


class CheckDist(Step):
    """Check dist."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Check dist")
        self.do_cmd("twine check dist/*")


class UploadToTestPyPI(Step):
    """Upload to test pypi."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Upload to TestPyPI")
        cmd = "twine upload --repository-url https://test.pypi.org/legacy/ dist/*"
        self.do_cmd(cmd)


class InstallFromTestPyPI(Step):
    """Install from test pypi."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Check installation from TestPyPI")
        self.do_cmd(
            f"sh build_tools/check_install_from_test_pypi.sh {context['version']}"
        )


class UploadToPyPI(Step):
    """Upload to pypi."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Upload to PyPI")
        self.do_cmd("twine upload dist/*")


class CheckVersionNumber(Step):
    """Check version number."""

    def action(self, context):
        """Carry out action."""
        self.instruct(
            f"Ensure that the following command gives version: {context['version']}"
        )
        p = context["package_name"]
        self.do_cmd(f'python -c "import {p}; print({p}.__version__)"')


class GitTagRelease(Step):
    """Git tag release."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Tag version as a release")
        self.do_cmd(f"git tag v{context['version']}")


class GitTagPreRelease(Step):
    """Git tag prerelease."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Tag version as a pre-release (increment as needed)")
        self.print_run(f"git tag v{context['version']}-rc.1")


class PushToGitHub(Step):
    """Push to GitHub."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Add and commit to git, then push to GitHub")


class CheckCIStatus(Step):
    """Check CI status."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Wait for CI to complete and check status")


class CheckOnlineDocs(Step):
    """Check online docs."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Check online docs")
        open_website(URLs.DOCS_ONLINE)


class CheckLocalDocs(Step):
    """Check local docs."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Check local docs")
        open_website(URLs.DOCS_LOCAL)


class CheckPyPIFiles(Step):
    """Check pypi files."""

    def action(self, context):
        """Carry out action."""
        self.instruct("Check PyPI files")
        open_website(URLs.PYPI)


class PushTagToGitHub(Step):
    """Push tag to GitHub."""

    def action(self, context):
        """Carry out action."""
        self.do_cmd(f"git push origin v{context['version']}")


def main():
    """Run release, main script."""
    colorama.init()
    steps = [
        # prepare and run final checks
        ConfirmGitStatus(branch="main"),
        MakeClean(),
        UpdateVersion(),
        CheckVersionNumber(),
        UpdateReadme(),
        MakeDocs(),
        CheckLocalDocs(),
        MakeDist(),
        CheckDist(),
        UploadToTestPyPI(),
        InstallFromTestPyPI(),
        PushToGitHub(),
        CheckCIStatus(),
        # make release
        GitTagRelease(),
        PushTagToGitHub(),
        CheckCIStatus(),
        CheckOnlineDocs(),
        UploadToPyPI(),
        CheckPyPIFiles(),
    ]
    context = dict()
    context["package_name"] = PACKAGE_NAME
    context["version"] = find_version(context["package_name"], "__init__.py")
    context["testdir"] = "temp/"
    for step in steps:
        step.run(context)
    cprint("\nDone!", color="yellow", style="bright")


if __name__ == "__main__":
    main()



"""Anomaly detection algorithms."""

from ._anomalisers import StatThresholdAnomaliser
from ._capa import CAPA
from ._circular_binseg import CircularBinarySegmentation

COLLECTIVE_ANOMALY_DETECTORS = [
    CAPA,
    CircularBinarySegmentation,
    StatThresholdAnomaliser,
]
ANOMALY_DETECTORS = COLLECTIVE_ANOMALY_DETECTORS

__all__ = ANOMALY_DETECTORS



"""Constant penalties for change and anomaly detection."""

import numpy as np

from ..utils.validation.parameters import check_larger_than


def make_bic_penalty(n_params: int, n: int, additional_cpts: int = 1) -> float:
    """Create a Bayesian Information Criterion (BIC) penalty.

    The BIC penalty is a constant penalty given by
    ``(n_params + additional_cpts) * log(n)``, where `n` is the sample size and
    `n_params` is the number of parameters per segment in the model across all
    variables, and `additional_cpts` is the number of additional change point parameters
    per segment. For change detection, this is 1.

    Parameters
    ----------
    n_params : int
        Number of model parameters per segment.
    n : int
        Sample size.
    additional_cpts: int, optional, default=1
        Number of additional change point parameters per segment. For change detection,
        this is 1.

    Returns
    -------
    float
        The BIC penalty value.
    """
    check_larger_than(1, n_params, "n_params")
    check_larger_than(1, n, "n")
    check_larger_than(0, additional_cpts, "additional_cpts")

    return (n_params + additional_cpts) * np.log(n)


def make_chi2_penalty(n_params: int, n: int) -> float:
    """Create a chi-square penalty.

    The penalty is the default penalty for the `CAPA` algorithm. It is described as
    "penalty regime 1" in the MVCAPA article [1]_. The penalty is based on a probability
    bound on the chi-squared distribution.

    The penalty is given by ``n_params + 2 * sqrt(n_params * log(n)) + 2 * log(n)``,
    where `n` is the sample size and `n_params` is the total number of parameters per
    segment in the model across all variables.

    Parameters
    ----------
    n_params : int
        Number of model parameters per segment.
    n : int
        Sample size.

    Returns
    -------
    float
        The chi-square penalty value.

    References
    ----------
    .. [1] Fisch, A. T., Eckley, I. A., & Fearnhead, P. (2022). Subset multivariate
       segment and point anomaly detection. Journal of Computational and Graphical
       Statistics, 31(2), 574-585.
    """
    check_larger_than(1, n_params, "n_params")
    check_larger_than(1, n, "n")

    psi = np.log(n)
    return n_params + 2 * np.sqrt(n_params * psi) + 2 * psi



"""Linear trend cost function.

This module contains the LinearTrendCost class, which is a cost function for
change point detection based on the squared error between data points
and a best fit linear trend line within each interval.
"""

import numpy as np

from ..base import BaseIntervalScorer
from ..utils.numba import njit
from ..utils.validation.parameters import check_data_column


@njit
def lin_reg_cont_piecewise_linear_trend_score(
    starts: np.ndarray,
    splits: np.ndarray,
    ends: np.ndarray,
    X: np.ndarray,
    times: np.ndarray,
) -> np.ndarray:
    """Evaluate the continuous linear trend cost.

    Parameters
    ----------
    starts : np.ndarray
        Start indices of the first intervals (inclusive).
    splits : np.ndarray
        Split indices between the intervals (contained in second interval).
    ends : np.ndarray
        End indices of the second intervals (exclusive).
    X : np.ndarray
        Data to evaluate. Must be a 2D array.
    times : np.ndarray
        Time steps corresponding to the data points. If the data points
        are evenly spaced, instead call the analytical score function.

    Returns
    -------
    scores : np.ndarray
        A 2D array of scores. One row for each interval. The number of columns
        is equal to the number of columns in the input data.
    """
    n_intervals = len(starts)
    n_columns = X.shape[1]
    scores = np.zeros((n_intervals, n_columns))

    for i in range(n_intervals):
        start, split, end = starts[i], splits[i], ends[i]
        split_interval_trend_data = np.zeros((end - start, 3))
        split_interval_trend_data[:, 0] = 1.0  # Intercept

        # Whole interval slope:
        split_interval_trend_data[:, 1] = times[start:end]  # Time steps

        # Change in slope from the 'split' index:
        split_interval_trend_data[(split - start) :, 2] = (
            times[split:end] - times[split]
        )

        ### Compute scores for all data columns at once:
        # Calculate the slope and intercept for the whole interval:
        split_interval_linreg_res = np.linalg.lstsq(
            split_interval_trend_data, X[start:end, :]
        )
        if (end - start) == 3 and split_interval_linreg_res[2] == 3:
            # If the interval is only 3 points long (the minimum),
            # and the model matrix is full rank, the residuals are zero.
            split_interval_squared_residuals = np.zeros((n_columns,))
        else:
            split_interval_squared_residuals = split_interval_linreg_res[1]

        # By only regressing onto the first two columns, we can calculate the cost
        # without allowing for a change in slope at the split point.
        joint_interval_linreg_res = np.linalg.lstsq(
            split_interval_trend_data[:, np.array([0, 1])], X[start:end, :]
        )
        joint_interval_squared_residuals = joint_interval_linreg_res[1]

        # If either of the linear regression solutions failed, return NaN.
        if (len(split_interval_squared_residuals) == 0) or (
            len(joint_interval_squared_residuals) == 0
        ):
            scores[i, :] = np.nan
        else:
            scores[i, :] = (
                joint_interval_squared_residuals - split_interval_squared_residuals
            )

    return scores


@njit
def continuous_piecewise_linear_trend_squared_contrast(
    signal: np.ndarray,
    first_interval_inclusive_start: int,
    second_interval_inclusive_start: int,
    non_inclusive_end: int,
):
    # Assume 'start' is the first index of the data, perform inner product with the
    # desired segment of the data to get the cost.
    assert (
        first_interval_inclusive_start
        < second_interval_inclusive_start
        < non_inclusive_end - 1
    )

    ## Translate named parameters to the NOT-paper sytax.
    ## We are zero-indexing the data, whilst the paper is one-indexing.
    s = first_interval_inclusive_start - 1
    e = non_inclusive_end - 1

    # Add one to NOT-syntax split index to account for the difference
    # in definition of where the change in slope starts from. They
    # let the change in slop take effect after the point before the split.
    b = second_interval_inclusive_start - 1 + 1

    l = e - s
    alpha = np.sqrt(
        6.0 / (l * (l**2 - 1) * (1 + (e - b + 1) * (b - s) + (e - b) * (b - s - 1)))
    )
    beta = np.sqrt(((e - b + 1.0) * (e - b)) / ((b - s - 1.0) * (b - s)))

    first_interval_slope = 3.0 * (b - s) + (e - b) - 1.0
    first_interval_constant = b * (e - s - 1.0) + 2.0 * (s + 1.0) * (b - s)

    second_interval_slope = 3.0 * (e - b) + (b - s) + 1.0
    second_interval_constant = b * (e - s - 1.0) + 2.0 * e * (e - b + 1)

    # Accumulate the contrast value inner product:
    contrast = 0.0
    for t in range(s + 1, b + 1):
        contrast += (
            alpha * beta * (first_interval_slope * t - first_interval_constant)
        ) * signal[t - first_interval_inclusive_start]

    for t in range(b + 1, e + 1):
        contrast += (
            (-alpha / beta) * (second_interval_slope * t - second_interval_constant)
        ) * signal[t - first_interval_inclusive_start]

    return np.square(contrast)


@njit
def analytical_cont_piecewise_linear_trend_score(
    starts: np.ndarray, splits: np.ndarray, ends: np.ndarray, X: np.ndarray
):
    """Evaluate the continuous piecewise linear trend cost.

    Using the analytical solution, this function evaluates the cost for
    `X[start:end]` for each each `[start, split, end]` triplett in `cuts`.

    Parameters
    ----------
    starts : np.ndarray
        Start indices of the first intervals (inclusive).
    splits : np.ndarray
        Split indices between the intervals (contained in second interval).
    ends : np.ndarray
        End indices of the second intervals (exclusive).
    X : np.ndarray
        Data to evaluate. Must be a 2D array.

    Returns
    -------
    scores : np.ndarray
        A 2D array of scores. One row for each interval. The number of columns
        is equal to the number of columns in the input data.
    """
    scores = np.zeros((len(starts), X.shape[1]))
    for i in range(len(starts)):
        start, split, end = starts[i], splits[i], ends[i]
        for j in range(X.shape[1]):
            scores[i, j] = continuous_piecewise_linear_trend_squared_contrast(
                X[start:end, j],
                first_interval_inclusive_start=start,
                second_interval_inclusive_start=split,
                non_inclusive_end=end,
            )

    return scores


class ContinuousLinearTrendScore(BaseIntervalScorer):
    """
    Continuous linear trend change score.

    Calculates the difference in squared error between observed data and:
    - a two-parameter linear trend across the whole interval, and
    - a three-parameter linear trend with a kink at the split point.

    Intended for use with the NOT segment selection method as developed by
    Baranowski et al. [1]_. Accessible within the `SeededBinarySegmentation`
    change detector by passing `selection_method="narrowest"`.

    By default, time steps are assumed to be evenly spaced.
    In this case, an analytical solution is used to calculate the score for each column
    in the data, as described in [1]_.

    If a time column is provided, its values are used as time stamps for calculating the
    linear trends. In this case, two linear regression problems are solved for each
    interval: One with a kink at the split point and one without.

    Parameters
    ----------
    time_column : str, optional
        Name of the time column in the data. The columns must be convertible to a float.
        If provided, its values are used as time stamps for calculating the piecewise
        linear trends.
        If not provided, the time steps are assumed to be evenly spaced.

    References
    ----------
    .. [1] Baranowski, R., Chen, Y., & Fryzlewicz, P. (2019). Narrowest-over-threshold
       detection of multiple change points and change-point-like features. Journal of
       the Royal Statistical Society Series B: Statistical Methodology, 81(3), 649-672.
    """

    _tags = {
        "authors": ["johannvk"],
        "maintainers": "johannvk",
        "task": "change_score",
    }

    def __init__(
        self,
        time_column: str | None = None,
    ):
        super().__init__()
        self.time_column = time_column
        self.time_column_idx = None
        self._time_stamps = None

    def _fit(self, X: np.ndarray, y=None):
        """Fit the cost.

        This method stores the input data for later cost evaluation.

        Parameters
        ----------
        X : np.ndarray
            Data to evaluate. Must be a 2D array.
        y: None
            Ignored. Included for API consistency by convention.
        """
        if self.time_column is not None:
            self.time_column_idx = check_data_column(
                self.time_column, "Time", X, self._X_columns
            )
        else:
            self.time_column_idx = None

        if self.time_column_idx is not None:
            # Need time column as float data for numba compatibility:
            self._time_stamps = X[:, self.time_column_idx].astype(np.float64)
            # Start at time zero for first data point:
            self._time_stamps -= self._time_stamps[0]
        else:
            # No provided time column or fixed parameters, so we assume
            # the time steps are [0, 1, 2, ..., n-1] for each segment.
            self._time_stamps = None

        if self.time_column_idx is not None:
            piecewise_linear_trend_columns = np.delete(
                np.arange(X.shape[1]), self.time_column_idx
            )
            self._piecewise_linear_trend_data = X[:, piecewise_linear_trend_columns]
        else:
            self._piecewise_linear_trend_data = X
        return self

    def _evaluate(self, cuts: np.ndarray) -> np.ndarray:
        """Evaluate the continuous piecewise linear trend scores.

        Evaluates the score on `X[start:end]` for each each `[start, split, end]`
        triplett in cuts.  On each interval, the difference in summed squared
        residuals between the best fit linear trend accross the whole interval
        and the best fit linear trend with a kink at the split point
        is calculated. The score is calculated for each column in the data.

        Parameters
        ----------
        cuts : np.ndarray
            A 2D array with three columns of integer locations.
            The first column is the ``start``, the second is the ``split``, and the
            third is the ``end`` of the interval to evaluate the score on.

        Returns
        -------
        scores : np.ndarray
            A 2D array of costs. One row for each interval. The number of
            columns is equal to the number of columns in the input data.
        """
        starts = cuts[:, 0]
        splits = cuts[:, 1]
        ends = cuts[:, 2]
        if self.time_column is None:
            scores = analytical_cont_piecewise_linear_trend_score(
                # scores = linear_trend_score(
                starts=starts,
                splits=splits,
                ends=ends,
                X=self._piecewise_linear_trend_data,
            )
        else:
            scores = lin_reg_cont_piecewise_linear_trend_score(
                starts=starts,
                splits=splits,
                ends=ends,
                X=self._piecewise_linear_trend_data,
                times=self._time_stamps,
            )
        return scores

    @property
    def min_size(self) -> int:
        """Minimum number of samples required on each side of a split point to evaluate.

        The size of each interval is defined as ``cuts[i, 1] - cuts[i, 0]``.
        To solve for a linear trend, we need at least 2 points on each side
        of the split point. Due to the construction of the score where we
        enfore continuity continuity at the split point, we need at least 1 point
        on the left side of the split point and 1 points on the right side of the
        split point. The minimum size of the total interval is therefore 3.

        Returns
        -------
        int
            The minimum valid size of an interval to evaluate.
        """
        # Need at least a difference of 1 between the start and split
        # indices to evaluate the cost, and at least a difference of 2
        # between the split and end indices to evaluate the cost.
        return 2

    def get_model_size(self, p: int) -> int:
        """Get the number of parameters in the cost function.

        Parameters
        ----------
        p : int
            Number of variables in the data.

        Returns
        -------
        int
            Number of parameters in the cost function.
        """
        # In each interval we need 2 parameters: slope and intercept.
        return 2 * p

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests.

        Returns
        -------
        params : list of dict
            Parameters to create testing instances of the class
        """
        params = [{}, {}]
        return params



"""The collective and point anomalies (CAPA) algorithm."""

__author__ = ["Tveten"]
__all__ = ["CAPA"]

import numpy as np
import pandas as pd

from ..anomaly_scores import L2Saving, to_saving
from ..base import BaseIntervalScorer
from ..compose.penalised_score import PenalisedScore
from ..penalties import make_chi2_penalty, make_linear_chi2_penalty
from ..utils.numba import njit
from ..utils.validation.data import check_data
from ..utils.validation.interval_scorer import check_interval_scorer
from ..utils.validation.parameters import check_larger_than, check_smaller_than
from ..utils.validation.penalties import check_penalty
from .base import BaseSegmentAnomalyDetector


@njit
def get_anomalies(
    anomaly_starts: np.ndarray,
) -> tuple[list[tuple[int, int]], list[tuple[int, int]]]:
    segment_anomalies = []
    point_anomalies = []
    i = anomaly_starts.size - 1
    while i >= 0:
        start_i = anomaly_starts[i]
        size = i - start_i + 1
        if size > 1:
            segment_anomalies.append((int(start_i), i + 1))
            i = int(start_i)
        elif size == 1:
            point_anomalies.append((i, i + 1))
        i -= 1
    return segment_anomalies, point_anomalies


def get_affected_components(
    penalised_scorer: PenalisedScore,
    anomalies: list[tuple[int, int]],
) -> list[tuple[int, int, np.ndarray]]:
    penalised_scorer.check_is_penalised()
    penalised_scorer.check_is_fitted()
    new_anomalies = []
    for start, end in anomalies:
        saving_values = penalised_scorer.score_.evaluate(np.array([start, end]))[0]
        saving_order = np.argsort(-saving_values)  # Decreasing order.
        penalised_savings = (
            np.cumsum(saving_values[saving_order]) - penalised_scorer.penalty_
        )
        argmax = np.argmax(penalised_savings)
        new_anomalies.append((start, end, saving_order[: argmax + 1]))
    return new_anomalies


def run_capa(
    segment_penalised_saving: PenalisedScore,
    point_penalised_saving: PenalisedScore,
    min_segment_length: int,
    max_segment_length: int,
    find_affected_components: bool = False,
) -> tuple[np.ndarray, list[tuple[int, int]], list[tuple[int, int]]]:
    segment_penalised_saving.check_is_penalised()
    segment_penalised_saving.check_is_fitted()
    point_penalised_saving.check_is_penalised()
    point_penalised_saving.check_is_fitted()
    n_samples = segment_penalised_saving.n_samples
    if not n_samples == point_penalised_saving.n_samples:
        raise ValueError(
            "The segment and point saving costs must span the same number of samples."
        )

    opt_savings = np.zeros(n_samples + 1)

    # Store the optimal start and affected components of an anomaly for each t.
    # Used to get the final set of anomalies after the loop.
    opt_anomaly_starts = np.repeat(np.nan, n_samples)
    starts = np.array([], dtype=int)
    max_segment_penalty = np.max(segment_penalised_saving.penalty_)

    ts = np.arange(min_segment_length - 1, n_samples)
    for t in ts:
        # Segment anomalies
        t_array = np.array([t])

        starts = np.concatenate((starts, t_array - min_segment_length + 1))
        ends = np.repeat(t + 1, len(starts))
        intervals = np.column_stack((starts, ends))
        segment_savings = segment_penalised_saving.evaluate(intervals).flatten()
        candidate_savings = opt_savings[starts] + segment_savings
        candidate_argmax = np.argmax(candidate_savings)
        opt_segment_saving = candidate_savings[candidate_argmax]
        opt_start = starts[0] + candidate_argmax

        # Point anomalies
        point_interval = np.column_stack((t_array, t_array + 1))
        point_savings = point_penalised_saving.evaluate(point_interval).flatten()
        opt_point_saving = opt_savings[t] + point_savings[0]

        # Combine and store results
        savings = np.array([opt_savings[t], opt_segment_saving, opt_point_saving])
        argmax = np.argmax(savings)
        opt_savings[t + 1] = savings[argmax]
        if argmax == 1:
            opt_anomaly_starts[t] = opt_start
        elif argmax == 2:
            opt_anomaly_starts[t] = t

        # Pruning the admissible starts
        saving_too_low = candidate_savings + max_segment_penalty <= opt_savings[t + 1]
        too_long_segment = starts < t - max_segment_length + 2
        prune = saving_too_low | too_long_segment
        starts = starts[~prune]

    segment_anomalies, point_anomalies = get_anomalies(opt_anomaly_starts)

    if find_affected_components:
        segment_anomalies = get_affected_components(
            segment_penalised_saving, segment_anomalies
        )
        point_anomalies = get_affected_components(
            point_penalised_saving,
            point_anomalies,
        )
    return opt_savings[1:], segment_anomalies, point_anomalies


def _make_chi2_penalty_from_score(score: BaseIntervalScorer) -> float:
    score.check_is_fitted()
    n = score.n_samples
    p = score.n_variables
    return make_chi2_penalty(score.get_model_size(p), n)


def _make_linear_chi2_penalty_from_score(score: BaseIntervalScorer) -> np.ndarray:
    score.check_is_fitted()
    n = score.n_samples
    p = score.n_variables
    return make_linear_chi2_penalty(score.get_model_size(p), n, p)


def _check_capa_penalised_score(score: BaseIntervalScorer, name: str, caller_name: str):
    if score.get_tag("is_penalised") and not isinstance(score, PenalisedScore):
        raise ValueError(
            f"{caller_name} only supports a penalised `{name}` constructed"
            " by `PenalisedScore`."
        )


class CAPA(BaseSegmentAnomalyDetector):
    """The collective and point anomaly (CAPA) detection algorithm.

    An efficient implementation of the CAPA family of algorithms for anomaly detection.
    Supports both univariate data [1]_ (CAPA) and multivariate data with subset
    anomalies [2]_ (MVCAPA) by using the penalised saving formulation of the collective
    anomaly detection problem found in [2]_ and [3]_. For multivariat data, the
    algorithm can also be used to infer the affected components for each anomaly given
    a suitable penalty array.

    Parameters
    ----------
    segment_saving : BaseIntervalScorer, optional, default=L2Saving()
        The saving to use for segment anomaly detection.
        If a cost is given, the saving is constructed from the cost. The
        cost must have a fixed parameter that represents the baseline cost.
        If a penalised saving is given, it must be constructed from `PenalisedScore`.
    segment_penalty : np.ndarray or float, optional, default=None
        The penalty to use for segment anomaly detection. If the segment penalty is
        penalised (`segment_penalty.get_tag("is_penalised")`) the penalty will
        be ignored. The different types of penalties are as follows:

        * ``float``: A constant penalty applied to the sum of scores across all
          variables in the data.
        * ``np.ndarray``: A penalty array of the same length as the number of
          columns in the data, where element ``i`` of the array is the penalty for
          ``i+1`` variables being affected by an anomaly. The penalty array
          must be positive and increasing (not strictly). A penalised score with a
          linear penalty array is faster to evaluate than a nonlinear penalty array.
        * ``None``: A default constant penalty is created in `predict` based on the
          fitted score using the `make_chi2_penalty` function.

    point_saving : BaseIntervalScorer, optional, default=L2Saving()
        The saving to use for point anomaly detection. Only savings with a
        minimum size of 1 are permitted.
        If a cost is given, the saving is constructed from the cost. The
        cost must have a fixed parameter that represents the baseline cost.
        If a penalised saving is given, it must be constructed from `PenalisedScore`.
    point_penalty : np.ndarray or float, optional, default=None
        The penalty to use for point anomaly detection. See the documentation for
        `segment_penalty` for details. For ``None`` input, the default is set using the
        `make_linear_chi2_penalty` function.
    min_segment_length : int, optional, default=2
        Minimum length of a segment. This may be overridden by the `min_size` of the
        fitted `segment_saving`.
    max_segment_length : int, optional, default=1000
        Maximum length of a segment.
    ignore_point_anomalies : bool, optional, default=False
        If ``True``, detected point anomalies are not returned by `predict`. I.e., only
        segment anomalies are returned. If ``False``, point anomalies are included in
        the output as segment anomalies of length 1.
    find_affected_components : bool, optional, default=False
        If ``True``, the affected components for each segment anomaly are returned in
        the `"icolumns"` key of the `predict` output.
        Only relevant for multivariate data in combination with a penalty array.
        The affected components are sorted from the highest to lowest evidence
        of an anomaly being present in the variable.

    References
    ----------
    .. [1] Fisch, A. T., Eckley, I. A., & Fearnhead, P. (2022). A linear time method
        for the detection of collective and point anomalies. Statistical Analysis and
        DataMining: The ASA Data Science Journal, 15(4), 494-508.

    .. [2] Fisch, A. T., Eckley, I. A., & Fearnhead, P. (2022). Subset multivariate
        collective and point anomaly detection. Journal of Computational and Graphical
        Statistics, 31(2), 574-585.

    .. [3] Tveten, M., Eckley, I. A., & Fearnhead, P. (2022). Scalable change-point and
        anomaly detection in cross-correlated data with an application to condition
        monitoring. The Annals of Applied Statistics, 16(2), 721-743.

    Examples
    --------
    >>> from skchange.anomaly_detectors import CAPA
    >>> from skchange.datasets import generate_piecewise_normal_data
    >>> df = generate_piecewise_normal_data(
    ...     means=[0, 10, 0, 20, 0],
    ...     lengths=[100, 20, 100, 10, 100],
    ...     seed=2,
    ... )
    >>> detector = CAPA()
    >>> detector.fit_predict(df)
            ilocs  labels
    0  [100, 120)       1
    1  [220, 230)       2
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": True,
    }

    def __init__(
        self,
        segment_saving: BaseIntervalScorer | None = None,
        segment_penalty: np.ndarray | float | None = None,
        point_saving: BaseIntervalScorer | None = None,
        point_penalty: np.ndarray | float | None = None,
        min_segment_length: int = 2,
        max_segment_length: int = 1000,
        ignore_point_anomalies: bool = False,
        find_affected_components: bool = False,
    ):
        self.segment_saving = segment_saving
        self.segment_penalty = segment_penalty
        self.point_saving = point_saving
        self.point_penalty = point_penalty
        self.min_segment_length = min_segment_length
        self.max_segment_length = max_segment_length
        self.ignore_point_anomalies = ignore_point_anomalies
        self.find_affected_components = find_affected_components
        super().__init__()

        _segment_score = L2Saving() if segment_saving is None else segment_saving
        check_interval_scorer(
            _segment_score,
            "segment_saving",
            "CAPA",
            required_tasks=["cost", "saving"],
            allow_penalised=True,
        )
        _segment_saving = to_saving(_segment_score)

        check_penalty(segment_penalty, "segment_penalty", "CAPA")
        if _segment_saving.get_tag("is_penalised"):
            _check_capa_penalised_score(_segment_saving, "segment_saving", "CAPA")
            self._segment_penalised_saving = _segment_saving.clone()
        else:
            self._segment_penalised_saving = PenalisedScore(
                _segment_saving,
                segment_penalty,
                make_default_penalty=_make_chi2_penalty_from_score,
            )
        _point_score = L2Saving() if point_saving is None else point_saving
        check_interval_scorer(
            _point_score,
            "point_saving",
            "CAPA",
            required_tasks=["cost", "saving"],
            allow_penalised=True,
        )
        if _point_score.min_size != 1:
            raise ValueError("`point_saving` must have `min_size == 1`.")
        _point_saving = to_saving(_point_score)

        check_penalty(point_penalty, "point_penalty", "CAPA")
        if _point_saving.get_tag("is_penalised"):
            _check_capa_penalised_score(_point_saving, "point_saving", "CAPA")
            self._point_penalised_saving = _point_saving.clone()
        else:
            self._point_penalised_saving = PenalisedScore(
                _point_saving,
                point_penalty,
                make_default_penalty=_make_linear_chi2_penalty_from_score,
            )

        check_larger_than(2, min_segment_length, "min_segment_length")
        check_larger_than(min_segment_length, max_segment_length, "max_segment_length")

        self.clone_tags(_segment_saving, ["distribution_type"])
        self.set_tags(
            **{"capability:variable_identification": find_affected_components}
        )

    def _predict(self, X: pd.DataFrame | pd.Series) -> pd.Series:
        """Detect events in test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame
            Time series to detect anomalies in.

        Returns
        -------
        y_sparse: pd.DataFrame
            A `pd.DataFrame` with a range index and two columns:
            * ``"ilocs"`` - left-closed ``pd.Interval``s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.
            * ``"icolumns"`` - list of affected columns for each segment anomaly. Only
              included if `find_affected_components` is ``True``.

        Attributes
        ----------
        fitted_segment_saving : BaseIntervalScorer
            The fitted penalised segment saving used for the detection.
        fitted_point_saving : BaseIntervalScorer
            The fitted penalised point saving used for the detection.
        scores : pd.Series
            The cumulative optimal savings for the input data.
        """
        X = check_data(
            X,
            min_length=self.min_segment_length,
            min_length_name="min_segment_length",
        )

        self.fitted_segment_saving: BaseIntervalScorer = (
            self._segment_penalised_saving.clone().fit(X)
        )
        self.fitted_point_saving: BaseIntervalScorer = (
            self._point_penalised_saving.clone().fit(X)
        )

        min_segment_length = max(
            self.min_segment_length, self.fitted_segment_saving.min_size
        )
        check_smaller_than(
            self.max_segment_length, min_segment_length, "min_segment_length"
        )
        check_larger_than(min_segment_length, X.shape[0], "X.shape[0]")
        opt_savings, segment_anomalies, point_anomalies = run_capa(
            segment_penalised_saving=self.fitted_segment_saving,
            point_penalised_saving=self.fitted_point_saving,
            min_segment_length=min_segment_length,
            max_segment_length=self.max_segment_length,
            find_affected_components=self.find_affected_components,
        )
        self.scores = pd.Series(opt_savings, index=X.index, name="score")

        anomalies = segment_anomalies
        if not self.ignore_point_anomalies:
            anomalies += point_anomalies
        anomalies = sorted(anomalies)

        return self._format_sparse_output(anomalies)

    def _transform_scores(self, X: pd.DataFrame | pd.Series) -> pd.DataFrame:
        """Return scores for predicted labels on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame, pd.Series or np.ndarray
            Data to score (time series).

        Returns
        -------
        scores : pd.DataFrame with same index as X
            Scores for sequence `X`.

        Notes
        -----
        The CAPA scores are the cumulative optimal savings, so the scores are increasing
        and are not per observation scores.
        """
        self.predict(X)
        return self.scores

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return ``"default"`` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skchange.anomaly_scores import L2Saving
        from skchange.base import BaseIntervalScorer
        from skchange.compose.penalised_score import PenalisedScore
        from skchange.costs import L2Cost
        from skchange.penalties import make_nonlinear_chi2_penalty

        def _make_nonlinear_chi2_penalty_from_score(
            score: BaseIntervalScorer,
        ) -> np.ndarray:
            score.check_is_fitted()
            n = score.n_samples
            p = score.n_variables
            return make_nonlinear_chi2_penalty(score.get_model_size(p), n, p)

        params = [
            {
                "segment_saving": PenalisedScore(
                    L2Saving(),
                    make_default_penalty=_make_nonlinear_chi2_penalty_from_score,
                ),
                "min_segment_length": 5,
                "max_segment_length": 100,
                "find_affected_components": True,
            },
            {
                "segment_saving": L2Cost(param=0.0),
                "point_saving": L2Cost(param=0.0),
                "min_segment_length": 5,
                "max_segment_length": 100,
            },
            {
                "segment_saving": L2Cost(param=0.0),
                "point_saving": L2Cost(param=0.0),
                "min_segment_length": 2,
                "max_segment_length": 20,
            },
        ]
        return params



"""Cost-based change scores."""

import numpy as np

from ..base import BaseIntervalScorer
from ..costs.base import BaseCost


def to_change_score(scorer: BaseIntervalScorer) -> BaseIntervalScorer:
    """Convert compatible scorers to a change score.

    Parameters
    ----------
    scorer : BaseIntervalScorer
        The scorer to convert to a change score. If a change score is provided, it is
        returned as is.

    Returns
    -------
    BaseIntervalScorer
        The change score.
    """
    if not isinstance(scorer, BaseIntervalScorer):
        raise ValueError(f"scorer must be a BaseIntervalScorer. Got {type(scorer)}.")
    task = scorer.get_tag("task")
    if task == "cost":
        change_score = ChangeScore(scorer)
    elif task == "change_score":
        change_score = scorer
    else:
        raise ValueError(
            f"scorer must have tag 'task' set to 'cost' or 'change_score'."
            f" Got scorer.get_tag('task') = {task}."
        )
    return change_score


class ChangeScore(BaseIntervalScorer):
    """Change score based on a cost class.

    The change score is calculated as cost difference under a no-change hypothesis
    versus a single change hypothesis.

    Parameters
    ----------
    cost : BaseCost
        The cost function.
    """

    _tags = {
        "authors": ["Tveten"],
        "maintainers": "Tveten",
        "task": "change_score",
    }

    def __init__(self, cost: BaseCost):
        self.cost = cost
        super().__init__()

        self.clone_tags(
            cost,
            ["distribution_type", "is_conditional", "is_aggregated", "is_penalised"],
        )

    @property
    def min_size(self) -> int | None:
        """Minimum valid size of an interval to evaluate."""
        if self.is_fitted:
            return self.cost_.min_size
        else:
            return self.cost.min_size

    def get_model_size(self, p: int) -> int:
        """Get the number of parameters to estimate over each interval.

        The primary use of this method is to determine an appropriate default penalty
        value in detectors.

        Parameters
        ----------
        p : int
            Number of variables in the data.
        """
        if self.is_fitted:
            return self.cost_.get_model_size(p)
        else:
            return self.cost.get_model_size(p)

    def _fit(self, X: np.ndarray, y=None):
        """Fit the change score.

        Parameters
        ----------
        X : np.ndarray
            Data to evaluate. Must be a 2D array.
        y : None
            Ignored. Included for API consistency by convention.

        Returns
        -------
        self :
            Reference to self.
        """
        self.cost_: BaseCost = self.cost.clone()
        self.cost_.fit(X)
        return self

    def _evaluate(self, cuts: np.ndarray) -> np.ndarray:
        """Evaluate the change score for a split within an interval.

        Parameters
        ----------
        cuts : np.ndarray
            A 2D array with three columns of integer locations.
            The first column is the ``start``, the second is the ``split``, and the
            third is the ``end`` of the interval to evaluate.
            The difference between subsets ``X[start:split]`` and ``X[split:end]`` is
            evaluated for each row in `cuts`.

        Returns
        -------
        scores : np.ndarray
            A 2D array of change scores. One row for each cut. The number of
            columns is 1 if the change score is inherently multivariate. The number of
            columns is equal to the number of columns in the input data if the score is
            univariate. In this case, each column represents the univariate score for
            the corresponding input data column.
        """
        left_intervals = cuts[:, [0, 1]]
        right_intervals = cuts[:, [1, 2]]
        full_intervals = cuts[:, [0, 2]]
        left_costs = self.cost_.evaluate(left_intervals)
        right_costs = self.cost_.evaluate(right_intervals)
        no_change_costs = self.cost_.evaluate(full_intervals)

        change_scores = no_change_costs - (left_costs + right_costs)

        # Ensure that slightly negative scores are set to 0:
        change_scores[(change_scores < 0) & (change_scores > -1e-8)] = 0.0

        return change_scores

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for interval scorers.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skchange.costs import GaussianCost, L2Cost

        params = [
            {"cost": L2Cost()},
            {"cost": GaussianCost()},
        ]
        return params



"""Interactive exploration of the PELT change detector."""

from skchange.change_detectors import PELT
from skchange.datasets import generate_alternating_data
from skchange.utils.benchmarking.profiler import Profiler

# Simple univariate example
df = generate_alternating_data(
    n_segments=2, mean=10, segment_length=100, p=1, random_state=2
)
detector = PELT(min_segment_length=1)
detector.fit(df)
detector.predict(df)
detector.transform(df)
detector.transform_scores(df)


# Profiling
n = int(1e6)
df = generate_alternating_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = PELT()
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()



"""Penalties and penalty functions for change and anomaly detection."""

from ._constant_penalties import (
    make_bic_penalty,
    make_chi2_penalty,
)
from ._linear_penalties import (
    make_linear_chi2_penalty,
    make_linear_penalty,
)
from ._nonlinear_penalties import make_mvcapa_penalty, make_nonlinear_chi2_penalty

CONSTANT_PENALTY_MAKERS = [
    make_bic_penalty,
    make_chi2_penalty,
]
LINEAR_PENALTY_MAKERS = [
    make_linear_penalty,
    make_linear_chi2_penalty,
]
NONLINEAR_PENALTY_MAKERS = [
    make_nonlinear_chi2_penalty,
    make_mvcapa_penalty,
]
PENALTY_MAKERS = [
    *CONSTANT_PENALTY_MAKERS,
    *LINEAR_PENALTY_MAKERS,
    *NONLINEAR_PENALTY_MAKERS,
]

__all__ = PENALTY_MAKERS
